# import bpy

import h5py
import numpy as np

# Start by loading the positions from file

#fi = h5py.File("../FullStriatumPos/network-neuron-positions.hdf5","r")
#fi = h5py.File("/home/hjorth/HBP/StriatumNetwork/model/Article-cube-2160/network-neuron-positions.hdf5","r")

# 2174 neurons, density 80500... 0.3x0.3x0.3
# hFile = "/home/hjorth/HBP/StriatumNetwork/model/Article2019HyperVoxel/network-neuron-positions.hdf5"
hFile = "/home/hjorth/HBP/Snudda/snudda/examples/networks/Net10062/network-neuron-positions.hdf5"
fi = h5py.File(hFile,"r")

keepAll = True
centreNeurons = True

xm = 0.005
ym = 0.0035
zm = 0.0027
wm = 0.0003

keepIdx = []

print("Loaded")

pos = fi["network/neurons/position"]
nColour = []

centre = np.mean(pos,axis=0)

for ir,row in enumerate(pos):
  if(ir % 50000 == 0):
    print(str(ir) + "/" + str(pos.shape[0]))
    
  if(keepAll or (xm <= row[0] and row[0] <= xm + wm \
                 and ym <= row[1] and row[1] <= ym + wm \
                 and zm <= row[2] and row[2] <= zm + wm)):
    keepIdx.append(ir)

    nType = fi["network/neurons/name"][ir].decode().split("_")[0]
    if(nType == "dSPN" or nType == "MSD1"):
      # col = [1.0, 0.2, 0.2]
      col = (77./255,151./255,1.0)
    elif(nType == "iSPN" or nType == "MSD2"):
      #col = [0.2, 0.2, 1.0]
      col = (67./255,55./255,181./255)
    elif(nType == "FS"):
      col = (6./255,31./255,85./255)
    elif(nType == "ChIN"):
      col = (252./266,102./255,0.0)
    elif(nType == "LTS"):
      col = (150./255,63./255,212./255)
    else:
      col = [0.4, 0.4, 0.4]
      
    nColour.append(col)

if(centreNeurons):
  insideCoords = pos[keepIdx,:] - centre
else:
  insideCoords = pos[keepIdx,:]
neuronColours = np.array(nColour)
posCol = np.concatenate((insideCoords,neuronColours),axis=1)
np.savetxt("hypervoxelcoords.txt",posCol,delimiter=",")

print("Wrote " + str(len(keepIdx)) + " neurons")

print("Next render... drawHyperCubeSomasRender.py")

#import pdb
#pdb.set_trace()
      



import unittest
import os

import numpy as np
from snudda.analyse.analyse_topology_activity import SnuddaAnalyseTopologyActivity


class TestTopologyAnalysis(unittest.TestCase):

    def setUp(self) -> None:
        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

    def test_basics(self):
        sa = SnuddaAnalyseTopologyActivity()

        with self.subTest("match_closest_unique"):

            spikes_a = np.array([1, 2, 3, 4, 5])
            spikes_b = np.array([1.001, 1.002, 1.995, 2.5, 4.020, 5])
            delta_t = 10e-3

            delta_a, delta_b = sa.match_closest_unique(spikes_a, spikes_b, delta_t=delta_t)

            idx_a = np.invert(np.isnan(delta_a))
            idx_b = np.invert(np.isnan(delta_b))

            ref_a = np.array([0.001, -0.005, np.nan, np.nan, 0])
            ref_b = np.array([-0.001, np.nan, 0.005, np.nan, np.nan, 0])

            self.assertTrue(np.allclose(delta_a[idx_a], ref_a[idx_a]))
            self.assertTrue((np.isnan(ref_a[np.isnan(delta_a)])).all())

            self.assertTrue(np.allclose(delta_b[idx_b], ref_b[idx_b]))
            self.assertTrue((np.isnan(ref_b[np.isnan(delta_b)])).all())

        with self.subTest("triggered"):

            spikes_a = np.array([1, 2, 3, 4, 5, 6])
            spikes_b = np.array([0, 0.1, 1.1, 1.2, 3.2, 4, 5.5, 10, 11])

            delta_t = sa.get_spike_triggered_deltas(spikes_a, spikes_b)

            self.assertTrue(np.isnan(delta_t[0:2]).all())
            self.assertTrue(np.allclose(delta_t[2:], np.array([0.1, 0.2, 0.2, 0, 0.5, 4, 5])))


if __name__ == '__main__':
    unittest.main()



import json

import numpy as np
from scipy.interpolate import griddata


class SnuddaRotate:
    """ Rotation object. """

    def __init__(self, config_file=None):

        """
        Constructor

        Args:
            config_file: Path to network config file
        """

        self.rotation_lookup = dict()
        self.config = None

        if config_file:
            self.parse_config_file(config_file=config_file)

    def parse_config_file(self, config_file):

        with open(config_file, "r") as f:
            self.config = json.load(f)

        # Parse the config
        for region_name, region_data in self.config["regions"].items():

            if "neuron_orientation" in region_data["volume"]:
                for neuron_type in region_data["volume"]["neuron_orientation"]:

                    orientation_info = region_data["volume"]["neuron_orientation"][neuron_type]
                    rotation_mode = orientation_info["rotation_mode"]
                    rotation_field_file = orientation_info["rotation_field_file"] if "rotation_field_file" in orientation_info else None
                    if rotation_field_file:
                        position, rotation = self.load_rotation_field(rotation_field_file, region_name)
                    else:
                        position, rotation = None, None

                    self.rotation_lookup[region_name, neuron_type] = (rotation_mode, position, rotation)

    @staticmethod
    def random_z_rotate(rng):
        """ Helper method, rotate around z-axis (of SWC coordinates)"""
        ang = 2 * np.pi * rng.uniform()

        return np.array([[np.cos(ang), -np.sin(ang), 0],
                         [np.sin(ang), np.cos(ang), 0],
                         [0, 0, 1]])

    def get_rotations(self, volume_name, neuron_type, neuron_positions, rng):
        """ Gets rotations for neuron_type in volumne name at neuron_positions """

        if (volume_name, neuron_type) in self.rotation_lookup:
            rotation_mode, field_position, field_rotation = self.rotation_lookup[volume_name, neuron_type]
        else:
            rotation_mode, field_position, field_rotation = "random", None, None

        if not rotation_mode or rotation_mode.lower() == "none":
            rotation_matrices = [np.eye] * neuron_positions.shape[0]

        elif rotation_mode in ["random", "default"]:
            rotation_matrices = [self.rand_rotation_matrix(rand_nums=rng.random(size=(3,)))
                                 for x in range(0, neuron_positions.shape[0])]

        elif "vector_field" in rotation_mode:
            rotation_vectors = griddata(points=field_position,
                                        values=field_rotation,
                                        xi=neuron_positions, method="linear")

            assert not np.isnan(np.sum(rotation_vectors)), \
                (f"Invalid rotation vector for volume {volume_name}, neuron {neuron_type}, "
                 f"is neuron position outside the field?\nNeuron positions: {neuron_positions}"
                 f" (must be inside convex hull of the field's positions points)")

            if rotation_mode == "vector_field_and_z":
                rotation_matrices = [np.matmul(SnuddaRotate.rotation_matrix_from_vectors(np.array([0, 0, 1]), rv),
                                               self.random_z_rotate(rng))
                                     for rv in rotation_vectors]
            else:
                # We need to rotate z-axis to point to rotation_vector
                rotation_matrices = [SnuddaRotate.rotation_matrix_from_vectors(np.array([0, 0, 1]), rv)
                                     for rv in rotation_vectors]
        else:
            raise TypeError(f"Unknown rotation mode {rotation_mode}")

        return rotation_matrices

    @staticmethod
    def rotation_matrix_from_vectors(vec1, vec2):
        """ Creates a rotation matrix to rotate vec1 into vec2."""

        # Special case, which gives cross product zero
        if (vec1 == vec2).all():
            return np.eye(3)

        # Taken from stackoverflow:
        # https://stackoverflow.com/questions/45142959/calculate-rotation-matrix-to-align-two-vectors-in-3d-space
        """ Find the rotation matrix that aligns vec1 to vec2
        :param vec1: A 3d "source" vector
        :param vec2: A 3d "destination" vector
        :return mat: A transform matrix (3x3) which when applied to vec1, aligns it with vec2.
        """
        a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
        v = np.cross(a, b)
        c = np.dot(a, b)
        s = np.linalg.norm(v)
        k_mat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        rotation_matrix = np.eye(3) + k_mat + k_mat.dot(k_mat) * ((1 - c) / (s ** 2))
        return rotation_matrix

    @staticmethod
    def test_rotation_matrix():

        vec1 = [2, 3, 2.5]
        vec2 = [-3, 1, -3.4]

        mat = SnuddaRotate.rotation_matrix_from_vectors(vec1, vec2)
        vec1_rot = mat.dot(vec1)
        assert np.allclose(vec1_rot / np.linalg.norm(vec1_rot), vec2 / np.linalg.norm(vec2))

    @staticmethod
    def load_rotation_field(rotation_field_file, volume_name):
        """ Loads rotation field for volumne_name from rotation_field_file """

        with open(rotation_field_file, "r") as f:
            rotation_field_data = json.load(f)

        if volume_name in rotation_field_data:
            assert "position" in rotation_field_data[volume_name] \
                   and "rotation" in rotation_field_data[volume_name], \
                f"Missing position and/or rotation tag in volume {volume_name}"

            return np.array(rotation_field_data[volume_name]["position"]) * 1e-6, \
                   np.array(rotation_field_data[volume_name]["rotation"])

        else:
            print(f"No volume name, assuming position and rotation is for {volume_name}")
            return np.array(rotation_field_data["position"]) * 1e-6, np.array(rotation_field_data["rotation"])

    @staticmethod
    def rand_rotation_matrix(deflection=1.0, rand_nums=None, rng=None):
        """
        Creates a random rotation matrix.

        deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
        rotation. Small deflection => small perturbation.
        rand_nums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
        """

        # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c

        if rand_nums is None:
            if rng is not None:
                rand_nums = rng.uniform(size=(3,))
            else:
                rand_nums = np.random.uniform(size=(3,))

        theta, phi, z = rand_nums

        theta = theta * 2.0 * deflection * np.pi  # Rotation about the pole (Z).
        phi = phi * 2.0 * np.pi  # For direction of pole deflection.
        z = z * 2.0 * deflection  # For magnitude of pole deflection.

        # Compute a vector V used for distributing points over the sphere
        # via the reflection I - V Transpose(V).  This formulation of V
        # will guarantee that if x[1] and x[2] are uniformly distributed,
        # the reflected points will be uniform on the sphere.  Note that V
        # has length sqrt(2) to eliminate the 2 in the Householder matrix.

        r = np.sqrt(z)
        vv = (np.sin(phi) * r, np.cos(phi) * r, np.sqrt(2.0 - z))

        st = np.sin(theta)
        ct = np.cos(theta)

        rr = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))

        # Construct the rotation matrix  ( V Transpose(V) - I ) R.
        mm = (np.outer(vv, vv) - np.eye(3)).dot(rr)
        return mm



from snudda.init.init import SnuddaInit
from collections import OrderedDict
import json
import os

if __name__ == "__main__":
  
  import argparse
  parser = argparse.ArgumentParser(description="Init custom network")
  parser.add_argument("network",type=str,help="Network path")
  parser.add_argument("level",choices=["0","1","2","3","4"],type=str,
                      help="Level of Parkinsson, 0=none, 4=bad")
  args = parser.parse_args()

  simName = args.network
  
  connectNeurons = False

  #simName = "networks/FSmorphTest2orig"
  #simName = "networks/FSmorphTest1b"
  #simName = "LTStest"
  #simName = "networks/twoFS"
  #simName = "networks/FSmorphTest4"
  #simName = "networks/3types-striatum"
  # simName = "networks/SynTest-v6" # MSMS tuning
  #simName = "networks/SynTest-v15"  

  neuronsDir = "cellspecs.parkinson/" + str(args.level) + "/"
  
  configName= simName + "/network-config.json"
  cnc = SnuddaInit(struct_def={}, config_file=configName, nChannels=1)
  cnc.define_striatum(num_dSPN=1500, num_iSPN=1500, num_FS=0, num_LTS=0, num_ChIN=0,
                      neurons_dir=neuronsDir,
                      volume_type="cube")

  # cnc.defineStriatum(nMSD1=0,nMSD2=0,nFS=100,nLTS=100,nChIN=0,volumeType="slice")



  dirName = os.path.dirname(configName)
  
  if not os.path.exists(dirName):
    os.makedirs(dirName)

  cnc.write_json(configName)

  if(not connectNeurons):
    print("Removing all target information, and rewriting config file")
    # Reopen the config file, and remove all connectivity settings, to
    # get an unconnected network

    with open(configName,"r") as f:
      conData = json.load(f,object_pairs_hook=OrderedDict)

    for k in conData:
      if(k.lower() in ["volume", "channels"]):
        continue

      # Remove targets
      if(False):
        x = conData[k]
        del x["targets"]
      
    with open(configName,"w") as f:
      print("Writing to file: " + str(configName))
      json.dump(conData,f,indent=2)
      

  print("Now run:\n./snudda.py place " + simName)
  print("./snudda.py detect " + simName)
  print("./snudda.py prune " + simName)
  print("./snudda.py input " + simName + " --input config/input-tinytest-v4.json")
  print("./snudda.py simulate " + simName \
        + " --voltOut " + simName + "/volt-out.csv --time 10.0")



import bpy
from bpy_extras.io_utils import ImportHelper
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy.types import Operator

def read_some_data(context, filepath):
    scale_f = 1000   # factor to downscale the data

    ''' read swc file '''
    #neuron_file = '/home/martin/Programs/Lmv5.2_32bit/entho.CNG.swc'
    print(filepath)
    f = open(filepath)
    lines = f.readlines()
    f.close()

    ''' find starting point '''
    x = 0
    while lines[x][0] is '#':
        x += 1
        
    ''' Create a dictionary with the first item '''
    data = lines[x].strip().split(' ')
    somaID = int(data[0])
    somaType = float(data[1])
    somaX = float(data[2])
    somaY = float(data[3])
    somaZ = float(data[4])
    somaR = float(data[5])
    somaParent = int(data[6])

    # We centre the neuron
    neuron = {somaID: [somaType,0.0,0.0,0.0,somaR,somaParent]}
    # neuron = {float(data[0]): [float(data[1]), float(data[2]), float(data[3]), float(data[4]), float(data[5]), float(data[6])]}
    x += 1
        
    ''' Read the rest of the lines to the dictionary '''
    for l in lines[x:]:
        data = l.strip().split(' ')

        compID = int(data[0])
        compType = float(data[1])
        compX = float(data[2])
        compY = float(data[3])
        compZ = float(data[4])
        compR = float(data[5])
        compParent = int(data[6])

        # Centre neuron, so soma is at 0,0,0
        neuron[compID] = [compType,
                          compX-somaX,
                          compY-somaY,
                          compZ-somaZ,
                          compR,
                          compParent]
        
        #neuron[float(data[0])] = [float(data[1]), float(data[2]), float(data[3]), float(data[4]), float(data[5]), float(data[6])]
        
    bpy.ops.object.empty_add(type='ARROWS', location=(neuron[1][1] / scale_f, neuron[1][2] / scale_f, neuron[1][3] / scale_f), rotation=(0, 0, 0))
    a = bpy.context.selected_objects[0]    
    a.name = 'neuron_swc'

    last = -10.0

    ''' Create object '''
    for key, value in neuron.items():


        if(value[0] == 1):
            # This is the soma, add it
            somaRadie = value[-2]
            bpy.ops.mesh.primitive_uv_sphere_add(location=(value[1] / scale_f ,
                                                           value[2] / scale_f ,
                                                           value[3] / scale_f),
                                                 size=somaRadie / scale_f)
            somaObj = bpy.context.selected_objects[0]
            somaObj.parent = a
            somaObj.select = False

            print("Adding soma " + str(value))
        
        if value[-1] == -1:
            continue
        
        if value[0] == 10:
            continue

        # if we need to start a new bezier curve
        if (value[-1] != last):
             # trace the origins
            tracer = bpy.data.curves.new('tracer','CURVE')
            tracer.dimensions = '3D'
            spline = tracer.splines.new('BEZIER')

            curve = bpy.data.objects.new('curve',tracer)
            curve.data.use_fill_caps = True # Added 2019-06-17
            bpy.context.scene.objects.link(curve)
            
            # render ready curve
            tracer.resolution_u = 8
            tracer.bevel_resolution = 8 # Set bevel resolution from Panel options
            tracer.fill_mode = 'FULL'
            tracer.bevel_depth = 1.0/scale_f # 0.001 # Set bevel depth from Panel options --- THIS REPLACES scale_f when setting radius
            
            # move nodes to objects
            p = spline.bezier_points[0]
            p.co = [neuron[value[-1]][1] / scale_f, neuron[value[-1]][2] / scale_f, neuron[value[-1]][3] / scale_f]
            # !!! Fixed a radie bug, was [5] should be [4] -- the first column is already removed /Johannes
            p.radius = neuron[value[-1]][4] #/ scale_f
            p.handle_right_type='VECTOR'
            p.handle_left_type='VECTOR'

            #import pdb
            #pdb.set_trace()
            
            if (last > 0):
                spline.bezier_points.add(1)            
                p = spline.bezier_points[-1]
                p.co = [value[1]/scale_f, value[2]/scale_f, value[3]/scale_f]
                #!!! Fixed a radie bug, was [5] should be [4]
                p.radius = value[4] #/ scale_f
                p.handle_right_type='VECTOR'
                p.handle_left_type='VECTOR'

            curve.parent = a
        
        # if we can continue the last bezier curve
        if value[-1] == last:
            spline.bezier_points.add(1)
            p = spline.bezier_points[-1]
            p.co = [value[1]/scale_f, value[2]/scale_f, value[3]/scale_f]
            #!!! Fixed a radie bug, was [5] should be [4]            
            p.radius = value[4] #/ scale_f
            p.handle_right_type='VECTOR'
            p.handle_left_type='VECTOR'
        
        last = key

    a.select = True
        
    return {'FINISHED'}

class ImportSWCData(Operator, ImportHelper):
    """This appears in the tooltip of the operator and in the generated docs"""
    bl_idname = "import_mesh.swc"  # important since its how bpy.ops.import_test.some_data is constructed
    bl_label = "Import SWC-Files"

    # ImportHelper mixin class uses this
    filename_ext = ".swc"

    filter_glob = StringProperty(
            default="*.swc",
            options={'HIDDEN'},
            )

    def execute(self, context):
        return read_some_data(context, self.filepath)


# Only needed if you want to add into a dynamic menu
def menu_func_import(self, context):
    self.layout.operator(ImportSWCData.bl_idname, text="SWC Importer")


def register():
    bpy.utils.register_class(ImportSWCData)
    bpy.types.INFO_MT_file_import.append(menu_func_import)


def unregister():
    bpy.utils.unregister_class(ImportSWCData)
    bpy.types.INFO_MT_file_import.remove(menu_func_import)


if __name__ == "__main__":
    register()



# This code assumes you have created a small network of neurons, it will
# setup current injections
#
#
# OBS! We set a holding current to keep neuron at around -80mV
# We also change GABA reversal potential to -40mV, since internal Cl is 30mM
# and external Cl is 135 mM, temperature is 306K
#
#
# How to:
#
# * Edit snudda_init_custom.py to have the neurons you want.
#
# * Generate network 
#
#  python3 snudda/simulate/network_pair_pulse_simulation.py setup Planert2010 networks/Planert2010-v1
#  snudda place networks/Planert2010-v1
#  snudda detect networks/Planert2010-v1
#  snudda prune networks/Planert2010-v1

# * Figure out where to put the slcie cut (with plotOnly equation is ignored)
# 
#  python3 snudda/utils/cut.py networks/Planert2010-v1/network-synapses.hdf5 "z>0" --plotOnly
#
# * Look at networks/Planert2010-v1/network-cut-slice.hdf5.pdf to decide cut plane
#
# * Compile mod files (we now have failure rates for GABA)
#
# nrnivmodl data/neurons/mechanisms/
#
# * Cut the slice, so z > 0.00504 is kept
#
#  python3 cut.py networks/Planert2010-v1/network-synapses.hdf5 "abs(z)<100e-6"
#
# * Look at networks/Planert2010-v1/network-cut-slice.hdf5.pdf to verify cut plane
#
# * Run dSPN -> iSPN calibration (you get dSPN -> dSPN data for free then)
#
#  mpiexec -n 12 -map-by socket:OVERSUBSCRIBE python3 snudda/simulate/network_pair_pulse_simulation.py run Planert2010 networks/Planert2010-v1/network-cut-slice.hdf5 --pre dSPN --post iSPN
#
# *  Analyse
#
#  python3 snudda/simulate/network_pair_pulse_simulation.py analyse networks/Planert2010-v1/network-cut-slice.hdf5 dSPN iSPN
#  python3 snudda/simulate/network_pair_pulse_simulation.py analyse Planert2010 networks/Planert2010-v1/network-cut-slice.hdf5 --pre dSPN --post dSPN
#
# * Look at plot with traces overlayed and histogram of voltage amplitudes
# (When you do preType to postType, you also get preType to preType for free
# since both voltages are recorded
import json
import os
import sys

import neuron
import numpy as np

from snudda.simulate.simulate import SnuddaSimulate
from snudda.utils import SnuddaLoadSimulation
from snudda.utils.load import SnuddaLoad
from snudda.utils import snudda_parse_path
from snudda.utils.snudda_path import get_snudda_data

from snudda.utils.ablate_network import SnuddaAblateNetwork

# We want to match Taverna 2008 data:

# The slices were 300 μm thick.  MSNs sampled were 25–100 μm from the
# surface of the slice to facilitate paired recordings.  Pairs always
# had cell bodies that were similar in depth (near the same focal
# plane) in an attempt to minimize the probability that the local axon
# collateral of one or the other cell was cut. After choosing a given
# pair of cells (lateral distance between somata, ≤50 μm)
#
# Assume loss of 100 micrometer in depth, at KI they start with 250 micrometers
# and get 150 micrometers after.

# TODO: Write code to check pair pulse ration


class SnuddaNetworkPairPulseSimulation:

    # TODO: Allow hold_voltage to be a dictionary, with neuron type as lookup
    #       to allow different holding voltages

    def __init__(self, network_path,
                 pre_type, post_type=None,
                 exp_type=None,
                 current_injection=10e-9,
                 current_injection_duration=1e-3,
                 n_stimulated_neurons=None, 
                 hold_voltage=-80e-3,
                 max_dist=50e-6,
                 log_file=None,
                 random_seed=None,
                 snudda_data=None):

        if os.path.isfile(network_path):
            self.network_file = network_path
            self.network_path = os.path.dirname(network_path)
        else:
            self.network_file = os.path.join(network_path, "network-synapses.hdf5")
            self.network_path = network_path

        self.snudda_data = get_snudda_data(snudda_data=snudda_data)

        self.exp_type = exp_type

        self.n_stimulated_neurons = n_stimulated_neurons

        self.pre_type = pre_type

        if post_type:
            self.post_type = post_type
        else:
            self.post_type = "ALL"

        self.cur_inj = current_injection
        self.hold_v = hold_voltage
        self.stim_all_at_once = False

        if log_file:
            self.log_file = log_file
        else:
            self.log_file = os.path.join(self.network_path, "log", "pair-pulse.log")

        print(f"Using log file {self.log_file}")

        self.max_dist = max_dist

        print(f"Checking depolarisation/hyperpolarisation of {pre_type} to {post_type} synapses")

        self.inj_spacing = 0.5  # Tried with 0.2 before, too close
        self.inj_duration = current_injection_duration

        self.snudda_sim = None  # Defined in run_sim
        self.snudda_load = None  # Defined in analyse
        self.data = None  # Defind in analyse
        self.holding_i_clamp_list = []
        self.pre_id = []
        self.possible_post_id = []
        self.inj_info = []
        self.exp_data = dict()
        self.random_seed = random_seed

    ############################################################################

    def setup(self, neuron_paths, neuron_names, number_of_neurons,
              connection_config=None, cut_outside=True,
              volume_type="slice",
              slice_depth = 150e-6,
              slice_side_len = 500e-6,
              random_seed=None, density=80500, d_min=15e-6,
              parallel=False,
              ipython_profile=None):

        if random_seed:
            self.random_seed = random_seed
        else:
            random_seed = self.random_seed

        if volume_type == "slice":
            side_len = slice_side_len
            # slice_depth = 150e-6  # now an inparameter
        else:
            side_len = None
            slice_depth = None

        from snudda import Snudda
        snd = Snudda(network_path=self.network_path, parallel=parallel, ipython_profile=ipython_profile)

        snd.init_tiny(neuron_paths=neuron_paths,
                      neuron_names=neuron_names,
                      number_of_neurons=number_of_neurons,
                      snudda_data=self.snudda_data,
                      connection_config=connection_config,
                      random_seed=random_seed,
                      volume_mesh=volume_type,
                      side_len=side_len,
                      slice_depth=slice_depth,
                      density=density,
                      d_min=d_min)

        snd.create_network()

        if cut_outside:
            from snudda.utils.cut import SnuddaCut
            cut = SnuddaCut(network_file=self.network_file,
                            cut_equation="(-75e-6 < z) & (z < 75e-6)",
                            show_plot=True)

            self.network_file = cut.out_file_name

    def ablate_post(self, pre_id=None, post_type=None, remove_pre_without_targets=False, include_gap_junctions=False):

        # This ablates neurons that are not connected to the pre synaptic neurons
        sa = SnuddaAblateNetwork(network_file=self.network_file)

        if pre_id:
            self.pre_id = pre_id

        elif self.pre_id is not None and len(self.pre_id) > 0:
            print(f"Using pre_id = {self.pre_id} (already defined)")
        else:
            self.pre_id = [x["neuron_id"] for x in sa.snudda_load.data["neurons"] if x["type"] == self.pre_type]
            if self.n_stimulated_neurons is not None:
                self.pre_id = self.pre_id[:self.n_stimulated_neurons]

        sa.keep_only_neurons_and_targets(neuron_id=self.pre_id, post_type=post_type,
                                         remove_pre_without_targets=remove_pre_without_targets,
                                         include_gap_junctions=include_gap_junctions)

        if remove_pre_without_targets:
            # We need to reduce the number of neurons stimulated, since we removed some of them
            self.pre_id = list(set(self.pre_id).intersection(set(sa.keep_neuron_id)))
            self.n_stimulated_neurons = len(self.pre_id)

            # Rerun filtering, this will also make sure we remove the pre neurons not stimulated
            sa.keep_only_neurons_and_targets(neuron_id=self.pre_id, post_type=post_type,
                                             remove_pre_without_targets=remove_pre_without_targets,
                                             include_gap_junctions=include_gap_junctions)

        new_network_file = self.network_file.replace(".hdf5", "") + "-ablated.hdf5"

        sa.write_network(out_file_name=new_network_file)

        # We need to remap neuron_id from the old to the new ablated network
        try:
            remap_dict = sa.get_remap_dictionary()
            self.pre_id = [remap_dict[int(x)] for x in self.pre_id]
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()


        self.network_file = new_network_file

    def setup_OLD(self, n_dSPN=120, n_iSPN=120, n_FS=20, n_LTS=0, n_ChIN=0,
              volume_type=None,
              neuron_density=80500,
              side_len=200e-6,
              slice_depth=150e-6,
              random_seed=None):

        """ Setup network for pair pulse simulation. If volume_type is 'slice', then side_len and slice_depth are used.
            If volume_type is 'cube' then side_len is used. If side_len is set to None then neuron_density is used.

        """

        from snudda.init.init import SnuddaInit

        if random_seed:
            self.random_seed = random_seed
        else:
            random_seed = self.random_seed

        if volume_type is None:
            volume_type = "slice"

        config_name = os.path.join(self.network_path, "network-config.json")
        cnc = SnuddaInit(struct_def={}, config_file=config_name, random_seed=self.random_seed,
                         snudda_data=self.snudda_data)
        cnc.define_striatum(num_dSPN=n_dSPN, num_iSPN=n_iSPN, num_FS=n_FS, num_LTS=n_LTS, num_ChIN=n_ChIN,
                            volume_type=volume_type, side_len=side_len, slice_depth=slice_depth,
                            neuron_density=neuron_density)

        dir_name = os.path.dirname(config_name)

        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        cnc.write_json(config_name)

        print(f"\n\nsnudda create {self.network_path}")
        print(f'python3 snudda/utils/cut.py {self.network_path}/network-synapses.hdf5 "abs(z)<100e-6"')

        print("\nThe last command will pop up a figure and enter debug mode,"
              " press ctrl+D in the terminal window after inspecting the plot to continue")

        print("\n!!! Remember to compile the mod files: nrnivmodl data/neurons/mechanisms")

        print("\nTo run for example dSPN -> iSPN (and dSPN->dSPN) calibration:")
        print(f"mpiexec -n 12 -map-by socket:OVERSUBSCRIBE python3 network_pair_pulse_simulation.py "
              f"run {self.exp_type} {self.network_path}/network-cut-slice.hdf5 dSPN iSPN")

        print(f"\npython3 snudda/simulate/network_pair_pulse_simulation.py analyse {self.exp_type} "
              f"{self.network_path}/network-cut-slice.hdf5 --pre dSPN --post iSPN"
              f"\npython3 snudda_network_pair_pulse_simulation.py analyse {self.network_path}/network-cut-slice.hdf5 "
              f"--pre iSPN --post dSPN")

    ############################################################################

    def write_simulation_config(self,
                                gaba_rev=None,
                                reversal_potential=None,
                                pre_id=None,
                                clamp_mode=None,
                                clamp_id=None,  # Override for neurons to clamp
                                stim_all_at_once=False,
                                return_run_str=False,
                                holding_current_init_time=0.1,
                                max_workers=None):

        if clamp_mode not in ("current", "voltage"):
            raise ValueError(f"Clamp mode {clamp_mode} is not supported. (use 'voltage' or 'current')")

        snudda_loader = SnuddaLoad(network_file=self.network_file)
        network_info = snudda_loader.data

        self.simulation_config_file = os.path.join(self.network_path, "simulation-config.json")

        if pre_id:
            print(f"Using user defined pre_id: {pre_id}")
            self.pre_id = pre_id
        elif self.pre_id:
            print(f"Using pre_id: {self.pre_id}")
        else:
            self.pre_id = [x["neuron_id"] for x in network_info["neurons"] if x["type"] == self.pre_type]
            if self.n_stimulated_neurons is not None:
                self.pre_id = self.pre_id[:self.n_stimulated_neurons]

        post_id = list(set([x["neuron_id"] for x in network_info["neurons"]]) - set(self.pre_id))

        self.inj_spacing = float(max(self.inj_spacing, holding_current_init_time))

        # inj_info contains (pre_id, inj_start_time)
        self.inj_info = list(zip(self.pre_id, self.inj_spacing + self.inj_spacing * np.arange(0, len(self.pre_id))))

        try:
            sim_end = self.inj_info[-1][1] + self.inj_spacing
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            print(e)
            import pdb
            pdb.set_trace()

        current_injection_info = dict()

        self.stim_all_at_once = stim_all_at_once

        if not stim_all_at_once:
            print(f"Sequential stimulation")
            # Sequential stimulation
            for pre_id, start_time in zip(self.pre_id, self.inj_spacing + self.inj_spacing * np.arange(0, len(self.pre_id))):
                start_time = float(start_time)
                current_injection_info[str(pre_id)] = { "time": [0, start_time, start_time + 1e-6,
                                                                 start_time + self.inj_duration,
                                                                 start_time + self.inj_duration + 1e-6,
                                                                 sim_end],
                                                        "current": [0, 0, self.cur_inj, self.cur_inj, 0, 0] }

        else:
            print("Stimulate all presynaptic neurons at once")
            # All at once stimulation
            sim_end = 2 * self.inj_spacing

            self.inj_info = [(self.pre_id, self.inj_spacing)]  # All presynaptic neurons are stimulated at t=self.inj_spacing

            for pre_id in self.pre_id:
                current_injection_info[str(pre_id)] = { "time": [0, self.inj_spacing, self.inj_spacing + 1e-6,
                                                                 self.inj_spacing + self.inj_duration,
                                                                 self.inj_spacing + self.inj_duration + 1e-6,
                                                                 sim_end],
                                                        "current": [0, 0, self.cur_inj, self.cur_inj, 0, 0] }

        if clamp_mode == "current":

            if clamp_id is None:
                clamp_id = post_id

            for p_id in clamp_id:
                current_injection_info[str(p_id)] = { "voltage": self.hold_v }

        if reversal_potential is None:
            reversal_potential = {}

        if gaba_rev is not None:
            if "ALL" not in reversal_potential:
                reversal_potential["ALL"] = dict()

            if "tmGabaA" in reversal_potential["ALL"]:
                raise ValueError("Reversal potential for tmGabaA already defined in 'reversal_potential'")

            reversal_potential["ALL"]["tmGabaA"] = gaba_rev


        sim_config = { "network_path": self.network_path,
                       "network_file": self.network_file,
                       "snudda_data": self.snudda_data,
                       "time": sim_end,
                       "log_file": "$network_path/log/output-log.txt",
                       "record_all_soma": True,
                       "current_injection_info": current_injection_info,
                       "reversal_potential_override": reversal_potential,  # {"ALL": {"tmGabaA": gaba_rev}},
                       "hold_voltage": self.hold_v,
                       "holding_current_init_time": holding_current_init_time
                       }

        if clamp_mode == "voltage":
            volt_clamp_info = dict()

            if clamp_id is None:
                clamp_id = post_id

            for p_id in clamp_id:
                volt_clamp_info[str(p_id)] = {"voltage": self.hold_v,
                                              "duration": sim_end,
                                              "save_current": True}

            sim_config["voltage_clamp"] = volt_clamp_info

        snudda_loader.close()

        print(f"Writing simulation config: {self.simulation_config_file}")
        try:
                with open(self.simulation_config_file, "w") as f:
                    json.dump(sim_config, f, indent=4)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            print(e)
            import pdb; pdb.set_trace()

        if max_workers is not None:
            worker_str = f"-n {max_workers}"
        else:
            worker_str = ""

        run_cmd = f"mpirun {worker_str} snudda simulate {self.network_path} --simulation_config {self.simulation_config_file}"

        if return_run_str:
            return run_cmd

    ############################################################################

    def setup_holding_volt(self, hold_v=None, sim_end=None):

        print("Starting setup_holding_volt...")

        assert sim_end is not None, "setup_holding_volt: Please set sim_end, for holding current"

        if hold_v is None:
            hold_v = self.hold_v

        if hold_v is None:
            print("Not using holding voltage, skipping.")
            return

        # Setup vClamps to calculate what holding current will be needed
        soma_v_clamp = []

        soma_list = [self.snudda_sim.neurons[x].icell.soma[0] for x in self.snudda_sim.neurons]

        for s in soma_list:
            vc = neuron.h.SEClamp(s(0.5))
            vc.rs = 1e-9
            vc.amp1 = hold_v * 1e3
            vc.dur1 = 100

            soma_v_clamp.append((s, vc))

        neuron.h.finitialize(hold_v * 1e3)
        neuron.h.tstop = 100
        neuron.h.run()

        self.holding_i_clamp_list = []

        # Setup iClamps
        for s, vc in soma_v_clamp:
            cur = float(vc.i)
            ic = neuron.h.IClamp(s(0.5))
            ic.amp = cur
            ic.dur = 2 * sim_end * 1e3
            self.holding_i_clamp_list.append(ic)

        # Remove vClamps
        v_clamps = None
        vc = None

        ############################################################################

    def set_gaba_rev(self, v_rev_cl):

        print(f"Setting GABA reversal potential to {v_rev_cl * 1e3} mV")

        for s_list in self.snudda_sim.synapse_dict.values():
            for s, _, _, _ in s_list:
                assert s.e == -60, "It should be GABA synapses only that we modify!"
                s.e = v_rev_cl * 1e3

    ############################################################################

    def pick_connected_pre_id(self, connection_type="synapses", post_type=None):

        if post_type is None:
            post_type = self.post_type

        if post_type == "ALL":
            print(f"pick_connected_pre_id requires post_type to be a specific neuron type. ('ALL' not valid here)")

        try:
            sl = SnuddaLoad(self.network_path)
            con_mat = sl.create_connection_matrix(sparse_matrix=False, connection_type=connection_type)

            pre_id = sl.get_neuron_id_of_type(neuron_type=self.pre_type)
            post_id = sl.get_neuron_id_of_type(neuron_type=post_type)

            n_con = np.sum(con_mat[pre_id, :][:, post_id], axis=1)
            possible_pre_id = pre_id[n_con > 0]
            np.random.shuffle(possible_pre_id)
            self.pre_id = possible_pre_id[:min(self.n_stimulated_neurons, len(possible_pre_id))]
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        print(f"Picking pre_id: {self.pre_id}")

    ############################################################################

    def run_sim(self, gaba_rev, pre_id=None, disable_gap_junctions=False, clamp_mode=None):

        self.snudda_sim = SnuddaSimulate(network_file=self.network_file,
                                         input_file=None,
                                         log_file=self.log_file,
                                         disable_gap_junctions=disable_gap_junctions)

        self.snudda_sim.setup()

        # A current pulse to all pre-synaptic neurons, one at a time
        if pre_id:
            print(f"Using user defined pre_id: {pre_id}")
            self.pre_id = pre_id
        elif self.pre_id is not None and len(self.pre_id) > 0:
            # pre_id already defined, good to go
            pass
        else:
            # pre_id not defined, set it up
            self.pre_id = [x["neuron_id"] for x in self.snudda_sim.network_info["neurons"] if x["type"] == self.pre_type]
            if self.n_stimulated_neurons is not None:
                self.pre_id = self.pre_id[:self.n_stimulated_neurons]

        # inj_info contains (pre_id, inj_start_time)
        self.inj_info = list(zip(self.pre_id, self.inj_spacing + self.inj_spacing * np.arange(0, len(self.pre_id))))

        sim_end = self.inj_info[-1][1] + self.inj_spacing

        print(f"Running with sim_end = {sim_end}s")

        # Set the holding voltage
        # self.setup_holding_volt(hold_v=self.hold_v, sim_end=sim_end)

        self.set_gaba_rev(gaba_rev)

        # Add current injections defined in init
        for (nid, t) in self.inj_info:
            print(f"Current injection to {nid} at {t} s")
            self.snudda_sim.add_current_injection(neuron_id=nid,
                                                  start_time=t,
                                                  end_time=t + self.inj_duration,
                                                  amplitude=self.cur_inj)

        # !!! We could maybe update code so that for postType == "ALL" we
        # record voltage from all neurons

        if clamp_mode == "voltage":
            if self.post_type == "ALL":
                post_id = list(set(self.snudda_sim.snudda_loader.get_neuron_id()) - set(self.pre_id))
                self.snudda_sim.add_voltage_clamp(cell_id=post_id, voltage=self.hold_v, duration=2*sim_end, save_i_flag=True)
            else:
                post_id = self.snudda_sim.snudda_loader.get_neuron_id_of_type(self.post_type)
                self.snudda_sim.add_voltage_clamp(cell_id=post_id, voltage=self.hold_v, duration=2*sim_end, save_i_flag=True)
            self.snudda_sim.add_volt_recording_soma()
        elif clamp_mode == "current":
            # Set the holding voltage
            self.setup_holding_volt(hold_v=self.hold_v, sim_end=sim_end)
            if self.post_type == "ALL":
                    self.snudda_sim.add_volt_recording_soma()
            else:
                # Record from all the potential post synaptic neurons
                post_id = self.snudda_sim.snudda_loader.get_neuron_id_of_type(self.post_type)

                # Also save the presynaptic traces for debugging, to make sure they spike
                pre_id = self.snudda_sim.snudda_loader.get_neuron_id_of_type(self.pre_type)

                id_to_record = set(pre_id).union(set(post_id))
                self.snudda_sim.add_volt_recording_soma(id_to_record)
        else:
            raise ValueError(f"Unknown clamp_mode = {clamp_mode}, should be 'voltage' or 'current'")

        # Run simulation
        self.snudda_sim.run(sim_end * 1e3, hold_v=self.hold_v)

        # Write results to disk
        # self.snudda_sim.write_voltage_OLD(self.volt_file)
        self.snudda_sim.write_output()

    ############################################################################

    # This extracts all the voltage deflections, to see how strong they are

    def analyse(self, max_dist=None, n_max_show=10, pre_id=None, post_type=None, clamp_mode=None, exp_data_file=None,
                force_post_id=None, time_window = 0.05, include_gap_junctions=False):

        import matplotlib
        import matplotlib.pyplot as plt

        model_mean = None
        model_std = None

        if clamp_mode == "voltage":
            data_unit = "current"
            plot_conversion = 1e9
            plot_unit = "nA"
            plot_unit_string = "Current"
        elif clamp_mode == "current":
            data_unit = "voltage"
            plot_conversion = 1e3
            plot_unit = "mV"
            plot_unit_string = "Voltage"
        else:
            raise ValueError(f"Unknown clamp_mode = {clamp_mode}, should be 'voltage' or 'current'")
        
        self.setup_exp_data(data_file=exp_data_file)

        if max_dist is None:
            max_dist = self.max_dist

        # Read the data
        self.snudda_load = SnuddaLoad(self.network_file)
        self.data = self.snudda_load.data

        ssd = SnuddaLoadSimulation(network_path=self.network_path)
        time = ssd.get_time()

        # Generate current info structure
        # A current pulse to all pre synaptic neurons, one at a time
        if pre_id:
            print(f"Using user provided pre_id = {pre_id}\n"
                  f"This must match what was used for simulation! BE CAREFUL!")
            self.pre_id = pre_id
        elif self.pre_id is not None:
            print(f"Using pre_id: {self.pre_id}\n")
        else:
            self.pre_id = [x["neuron_id"] for x in self.data["neurons"] if x["type"] == self.pre_type]
            if self.n_stimulated_neurons is not None:
                self.pre_id = self.pre_id[:self.n_stimulated_neurons]

        if post_type is None:
            post_type = self.post_type

        assert post_type != "ALL", "You need to specify a neuron type as post_type, e.g. FS"

        assert self.post_type == post_type or self.post_type == "ALL", \
            f"You can only analyse post_type data that you recorded (e.g. {self.post_type}), " \
            f"to record data from all neuron types use post_type=ALL"

        self.possible_post_id = [x["neuron_id"] for x in self.data["neurons"] if x["type"] == post_type]
        self.possible_post_id = list(set(self.possible_post_id)-set(self.pre_id))

        recorded_data = ssd.get_data(data_unit, self.possible_post_id)[0]

        # injInfo contains (preID,injStartTime)
        if self.stim_all_at_once:
            self.inj_info = [(self.pre_id, self.inj_spacing)]
        else:
            self.inj_info = zip(self.pre_id, self.inj_spacing + self.inj_spacing * np.arange(0, len(self.pre_id)))

        # For each pre synaptic neuron, find the voltage deflection in each
        # of its post synaptic neurons

        synapse_data = []
        too_far_away = 0

        con_mat = self.snudda_load.create_connection_matrix(sparse_matrix=False)

        if include_gap_junctions:
            con_mat_gj = self.snudda_load.create_connection_matrix(sparse_matrix=False, connection_type="gap_junctions")
            con_mat = np.abs(con_mat) + np.abs(con_mat_gj)

        for (pre_id, t) in self.inj_info:
            # Post synaptic neuron to preID
            if isinstance(pre_id, list):
                post_id_set = set(np.where(np.sum(con_mat[pre_id, :], axis=1))[0])
            else:
                post_id_set = set(np.where(con_mat[pre_id,:])[0])

            post_id_set = post_id_set.intersection(self.possible_post_id)

            pre_pos = self.snudda_load.data["neuron_positions"][pre_id, :]

            if force_post_id is not None:
                print(f"Forcing post id {force_post_id}")
                post_id_set = force_post_id

            for post_id in post_id_set:

                if max_dist is not None:
                    post_pos = self.snudda_load.data["neuron_positions"][post_id, :]
                    if np.linalg.norm(pre_pos - post_pos) > max_dist:
                        too_far_away += 1
                        continue

                # There is a bit of synaptic delay, so we can take voltage
                # at first timestep as baseline

                if t + time_window > np.max(time):
                    print(f"Simulation only run to {np.max(time)}s, missing pulses at {t}s (check_width={time_window}s)")
                    continue

                t_idx = np.where(np.logical_and(t <= time, time <= t + time_window))[0]

                if post_id not in recorded_data:
                    print(f"Missing key {post_id}")
                    import pdb
                    pdb.set_trace()

                try:
                    synapse_data.append((time[t_idx], recorded_data[post_id][t_idx].flatten(), pre_id, post_id))

                    assert len(t_idx) > 0, f"Internal error, no time points recorded between {t} and {t + time_window} " \
                                           f"for synapse pre_id={pre_id}, post_id={post_id}"
                except:
                    import traceback
                    print(traceback.format_exc())
                    import pdb
                    pdb.set_trace()


        if max_dist is not None:
            print(f"Number of pairs excluded, distance > {max_dist * 1e6} mum : {too_far_away}")

        # Fig names:
        trace_fig = os.path.join(os.path.dirname(self.network_file),
                                 "figures",
                                 f"{self.exp_type}-synapse-calibration-{clamp_mode}-clamp-traces-{self.pre_type}-{post_type}.pdf")

        hist_fig = os.path.join(os.path.dirname(self.network_file),
                                "figures",
                                f"{self.exp_type}-synapse-calibration-{clamp_mode}-clamp-histogram-{self.pre_type}-{post_type}.pdf")

        fig_dir = os.path.join(os.path.dirname(self.network_file), "figures")

        if not os.path.exists(fig_dir):
            os.makedirs(fig_dir)

        # Extract the amplitude of all voltage pulses
        amp = np.zeros((len(synapse_data),))
        idx_max = np.zeros((len(synapse_data),), dtype=int)
        t_max = np.zeros((len(synapse_data),))

        # Save the largest deflection -- with sign
        try:
            for i, (t, data, pre_id, post_id) in enumerate(synapse_data):
                idx_max[i] = np.argmax(np.abs(data - data[0]))
                t_max[i] = t[idx_max[i]] - t[0]
                amp[i] = data[idx_max[i]] - data[0]
        
        except:
            import traceback
            t_str = traceback.format_exc()
            print(t_str)
            import pdb
            pdb.set_trace()

        if len(amp) <= 0:
            print("No responses... too short distance!")
            return None, None, None, None 

        print(f"Min amp: {np.min(amp)}")
        print(f"Max amp: {np.max(amp)}")
        print(f"Mean amp: {np.mean(amp)} +/- {np.std(amp)}")
        print(f"Amps: {amp}")

        # Now we have all synapse deflections in synapseData
        matplotlib.rcParams.update({'font.size': 22})

        sort_idx = np.argsort(amp)
        if n_max_show is not None and len(sort_idx) > n_max_show:
            keep_idx = [sort_idx[int(np.round(x))] for x in np.linspace(0, len(sort_idx) - 1, n_max_show)]
        else:
            keep_idx = sort_idx

        plt.figure()
        
        for x in keep_idx:
            t, data, pre_id, post_id = synapse_data[x]
            plt.plot((t - t[0]) * 1e3, (data - data[0]) * plot_conversion, color="black")

        plt.scatter(t_max * 1e3, amp * plot_conversion, color="blue", marker=".", s=100)

        if (self.exp_type, self.pre_type, post_type) in self.exp_data:
            exp_mean, exp_std = self.exp_data[(self.exp_type, self.pre_type, post_type)]

            t_end = (t[-1] - t[0]) * 1e3

            axes = plt.gca()
            ay = axes.get_ylim()
            # Plot SD or 1.96 SD?
            plt.errorbar(t_end, exp_mean * plot_conversion, exp_std * plot_conversion, ecolor="red",
                         marker='o', color="red")
            
            model_mean = np.mean(amp) * plot_conversion
            model_std = np.std(amp) * plot_conversion
            plt.errorbar(t_end - 2, model_mean, model_std, ecolor="blue",
                         marker="o", color="blue")

            axes.set_ylim(ay)

        plt.xlabel("Time (ms)")
        plt.ylabel(f"{plot_unit_string} ({plot_unit})") 
        # plt.title(str(len(synapseData)) + " traces")
        plt.title(f"{self.pre_type} to {post_type}")

        # Remove part of the frame
        plt.gca().spines["right"].set_visible(False)
        plt.gca().spines["top"].set_visible(False)

        plt.tight_layout()
        plt.ion()
        plt.savefig(trace_fig, dpi=300)
        plt.show()

        plt.figure()
        plt.hist(amp * plot_conversion, bins=20)
        plt.title(f"{self.pre_type} to {post_type}")
        plt.xlabel(f"{plot_unit_string} deflection ({plot_unit})")

        # Remove part of the frame
        plt.gca().spines["right"].set_visible(False)
        plt.gca().spines["top"].set_visible(False)

        plt.tight_layout()
        plt.savefig(hist_fig, dpi=300)
        plt.show()

        plt.pause(5)

        return model_mean, model_std, trace_fig, hist_fig

    ############################################################################

    def fraction_of_pop_responding_at_time(self, neuron_type, clamp_mode, time, time_window, min_change):

        if clamp_mode == "voltage":
            data_unit = "current"
        elif clamp_mode == "current":
            data_unit = "voltage"
        else:
            raise ValueError(f"Unknown clamp_mode = {clamp_mode}, should be 'voltage' or 'current'")


        post_id = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type)

        ssd = SnuddaLoadSimulation(network_path=self.network_path)
        t = ssd.get_time()

        recorded_data = ssd.get_data(data_unit, post_id)[0]

        time_idx = np.where(np.logical_and(time < t, t < time + time_window))

        above_threshold_ctr = 0
        total_ctr = 0

        for p_id in post_id:
            data = recorded_data[p_id][time_idx].flatten()
            if np.abs(np.max(data) - np.min(data)) > min_change:
                above_threshold_ctr += 1

            total_ctr += 1

        print(f"{above_threshold_ctr} out of {total_ctr} respond at between {time} and {time + time_window}s, {above_threshold_ctr/float(total_ctr)*100:.2f} %")

    ############################################################################

    def setup_exp_data(self, data_file=None):

        self.exp_data = dict()

        if data_file is None:
            data_file = snudda_parse_path(os.path.join("$DATA", "synapses", "pair_pulse_experiment_data.json"),
                                          self.snudda_data)

        with open(data_file, "r") as f:
            exp_data = json.load(f)

        for exp_name, exp_info in exp_data.items():
            for pre_neuron in exp_info:
                for post_neuron in exp_info[pre_neuron]:
                    self.exp_data[exp_name, pre_neuron, post_neuron] = exp_info[pre_neuron][post_neuron]

    ############################################################################

    def analyse_gap_junctions(self, clamp_mode, exp_data_file=None, pre_id=None, post_type=None,
                              time_window=0.05, baseline_offset=-0.05):

        import matplotlib.pyplot as plt
        import seaborn as sns

        model_mean = None
        model_std = None

        if clamp_mode == "voltage":
            data_unit = "current"
            plot_conversion = 1e9
            plot_unit = "nA"
            plot_unit_string = "Current"
        elif clamp_mode == "current":
            data_unit = "voltage"
            plot_conversion = 1e3
            plot_unit = "mV"
            plot_unit_string = "Voltage"
        else:
            raise ValueError(f"Unknown clamp_mode = {clamp_mode}, should be 'voltage' or 'current'")

        self.setup_exp_data(data_file=exp_data_file)

        # Read the data
        self.snudda_load = SnuddaLoad(self.network_file)
        print(f"Loading {self.network_file}")
        self.data = self.snudda_load.data

        ssd = SnuddaLoadSimulation(network_path=self.network_path)
        time = ssd.get_time()

        if pre_id:
            print(f"Using user provided pre_id = {pre_id}\n"
                  f"This must match what was used for simulation! BE CAREFUL!")
            self.pre_id = pre_id
        elif self.pre_id is not None:
            print(f"Using pre_id: {self.pre_id}\n")
        else:
            self.pre_id = [x["neuron_id"] for x in self.data["neurons"] if x["type"] == self.pre_type]
            if self.n_stimulated_neurons is not None:
                self.pre_id = self.pre_id[:self.n_stimulated_neurons]

        if post_type is None:
            post_type = self.post_type

        assert post_type != "ALL", "You need to specify a neuron type as post_type, e.g. FS"

        assert self.post_type == post_type or self.post_type == "ALL", \
            f"You can only analyse post_type data that you recorded (e.g. {self.post_type}), " \
            f"to record data from all neuron types use post_type=ALL"

        self.possible_post_id = [x["neuron_id"] for x in self.data["neurons"] if x["type"] == post_type]
        self.possible_post_id = list(set(self.possible_post_id)-set(self.pre_id))

        if len(self.possible_post_id) == 0:
            print(f"No post neurons. Did you pick the wrong {post_type =}")

        recorded_data = ssd.get_data(data_unit, self.possible_post_id)[0]
        recorded_data_pre = ssd.get_data(data_unit, self.pre_id)[0]

        if self.stim_all_at_once:
            self.inj_info = [(self.pre_id, self.inj_spacing)]
        else:
            self.inj_info = zip(self.pre_id, self.inj_spacing + self.inj_spacing * np.arange(0, len(self.pre_id)))

        synapse_data = []

        con_mat_gj = self.snudda_load.create_connection_matrix(sparse_matrix=False, connection_type="gap_junctions")

        coupling_coefficient = []

        for (pre_id, t) in self.inj_info:

            if t + time_window > np.max(time):
                print(f"Simulation only run to {np.max(time)}s, missing pulses at {t}s (check_width={window_width}s)")
                continue

            t_idx = np.where(np.logical_and(t <= time, time <= t + time_window))[0]

            baseline_idx = np.where(np.logical_and(t + baseline_offset <= time,
                                                   time <= t + baseline_offset + time_window))[0]

            pre_base = np.mean(recorded_data_pre[pre_id][baseline_idx].flatten())
            pre_amplitude = np.max(np.abs(recorded_data_pre[pre_id][t_idx].flatten() - pre_base))

            # Post synaptic neuron to preID
            if isinstance(pre_id, list):
                post_id_set = set(np.where(np.sum(con_mat_gj[pre_id, :], axis=1))[0])
            else:
                post_id_set = set(np.where(con_mat_gj[pre_id,:])[0])

            post_id_set = post_id_set.intersection(self.possible_post_id)

            for post_id in post_id_set:

                if post_id not in recorded_data:
                    print(f"Missing key {post_id}")
                    import pdb
                    pdb.set_trace()

                try:
                    synapse_data.append((time[t_idx], recorded_data[post_id][t_idx].flatten(), pre_id, post_id))

                    assert len(t_idx) > 0, f"Internal error, no time points recorded between {t} and {t + time_window} " \
                                           f"for synapse pre_id={pre_id}, post_id={post_id}"
                except:
                    import traceback
                    print(traceback.format_exc())
                    import pdb
                    pdb.set_trace()

                try:
                    post_base = np.mean(recorded_data[post_id][baseline_idx].flatten())
                    post_amplitude = np.max(np.abs(recorded_data[post_id][t_idx].flatten() - post_base))

                    coupling_coefficient.append(post_amplitude / pre_amplitude)
                except Exception as e:
                    import traceback
                    print(traceback.format_exc())
                    print(e)
                    import pdb
                    pdb.set_trace()

        print(f"Mean coupling coefficient: {np.mean(coupling_coefficient)} +/- {np.std(coupling_coefficient)}")
        print(coupling_coefficient)

        sns.violinplot(y=coupling_coefficient, inner="box")

        plt.ylabel("Coupling coefficient")
        plt.title("Gap Junction Coupling")

    ############################################################################


if __name__ == "__main__":

    if '-python' in sys.argv:
        print("Called through NEURON special file, fixing arguments")
        pythonidx = sys.argv.index('-python')
        if len(sys.argv) > pythonidx:
            sys.argv = sys.argv[pythonidx + 1:]
    
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Calibrate synapse conductances")
    parser.add_argument("task", choices=["setup", "run", "analyse"])
    parser.add_argument("expType", help="Experiment we replicate", choices=["Planert2010", "Szydlowski2013"])
    parser.add_argument("networkFile", help="Network file (hdf5) or network directory")
    parser.add_argument("--preType", "--pre", help="Pre synaptic neuron type", default="dSPN")
    parser.add_argument("--postType", "--post", default="ALL",
                        help="Post synaptic neuron type (for run task, "
                             "postType can be 'ALL' to record from all neuron)")
    parser.add_argument("--maxDist", help="Only check neuron pairs within (mum)", type=float, default=None)
    parser.add_argument("--nShow", help="Number of traces to show", type=int, default=0)
    args = parser.parse_args()

    if args.maxDist is None:
        max_dist = 50e-6
    elif args.maxDist == "None":
        max_dist = None
    else:
        max_dist = float(args.maxDist)

    print(f"Using maxDist = {max_dist}")

    if args.expType == "Planert2010":
        n_dSPN = 120
        n_iSPN = 120
        n_FS = 20
        n_LTS = 0
        n_ChIN = 0

        hold_v = -80e-3
        max_dist = 100e-6 if args.maxDist is None else args.maxDist
        GABA_rev = -40e-3

    elif args.expType == "Szydlowski2013":
        n_dSPN = 10
        n_iSPN = 10
        n_FS = 20
        n_LTS = 20
        n_ChIN = 0

        hold_v = -76e-3
        max_dist = 200e-6 if args.maxDist is None else args.maxDist
        GABA_rev = -39e-3

    else:
        print(f"Unknown expType = {args.expType}")
        sys.exit(-1)

    pps = SnuddaNetworkPairPulseSimulation(network_path=args.networkFile,
                                           exp_type=args.expType,
                                           pre_type=args.preType,
                                           post_type=args.postType,
                                           max_dist=max_dist,
                                           hold_voltage=hold_v,
                                           clamp_mode="current")

    if args.task == "setup":
        pps.setup(n_dSPN=n_dSPN, n_iSPN=n_iSPN,
                  n_FS=n_FS, n_LTS=n_LTS, n_ChIN=n_ChIN)

    elif args.task == "run":
        pps.run_sim(gaba_rev=GABA_rev)

    elif args.task == "analyse":

        if args.nShow == 0:
            n_show = None
        else:
            n_show = args.nShow

        pps.analyse(n_max_show=n_show)



import numpy as np
from collections import OrderedDict
import json
from snudda.utils.numpy_encoder import NumpyEncoder

volume_name = "Striatum"

file_lookup = { "LTS" : "1001_STRd_Sst_dens_smooth.csv",
                "ChIN" : "252_STRd_Chat_dens_smooth.csv",
                "dSPN" : "352_STRd_Drd1_dens_smooth.csv",
                "iSPN" : "72109410_STRd_Adora2a_dens_smooth.csv",
                "FS" : "79556738_STRd_Pvalb_dens_smooth.csv"}

output_file = "dorsal_striatum_density.json"

density_data = OrderedDict()

for neuron_type, density_file in file_lookup.items():
    print(f"Loading {neuron_type} data from {density_file}")
    file_data = np.genfromtxt(density_file, delimiter=' ')
    
    coord = file_data[:,:3]
    density = file_data[:,3]

    density_data[neuron_type] = OrderedDict()
    density_data[neuron_type]["Coordinates"] = coord
    density_data[neuron_type]["Density"] = density

volume_density_data = OrderedDict()
volume_density_data[volume_name] = density_data
    
with open(output_file, "w") as f:
    json.dump(volume_density_data, f, indent=4, cls=NumpyEncoder)

print(f"Wrote {output_file}")



import unittest
import os
import sys
import numpy as np

from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.detect.detect import SnuddaDetect
from snudda.place.place import SnuddaPlace

from scipy.spatial.distance import pdist

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))


class TestDetectSynapseCluster(unittest.TestCase):

    """ This tests the cluster generation by synapse duplication. """

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_detect_cluster")

        create_cube_mesh(file_name=os.path.join(network_path, "mesh", "simple_mesh.obj"),
                         centre_point=(0, 0, 0),
                         side_len=500e-6)

        config_file = os.path.join(network_path, "network-config.json")
        position_file = os.path.join(network_path, "network-neuron-positions.hdf5")
        save_file = os.path.join(network_path, "voxels", "network-putative-synapses.hdf5")

        #  TODO: If d_view is None code run sin serial, add test parallel
        sp = SnuddaPlace(config_file=config_file, d_view=None, verbose=True)
        sp.place()

        # We want to load in the ball and stick neuron that has 20 micrometer soma diameter, and axon (along y-axis),
        # and dendrite along (x-axis) out to 100 micrometer distance from centre of soma.

        self.sd = SnuddaDetect(config_file=config_file, position_file=position_file,
                               save_file=save_file, rc=None,
                               hyper_voxel_size=150, verbose=True)

    def test_clustering(self):

        neuron_positions = np.array([[0, 20, 0],    # Postsynaptiska
                                     [0, 80, 0],
                                     [0, 140, 0],
                                     [0, 200, 0],
                                     [20, 0, 0],    # Presynaptiska
                                     [80, 0, 0],
                                     [140, 0, 0],
                                     [200, 0, 0]
                                     ])*1e-6

        for idx, pos in enumerate(neuron_positions):
            self.sd.neurons[idx]["position"] = pos

        ang = -np.pi/2
        R_x = np.array([[1, 0, 0],
                        [0, np.cos(ang), -np.sin(ang)],
                        [0, np.sin(ang), np.cos(ang)]])

        ang = np.pi/2
        R_y = np.array([[np.cos(ang), 0, np.sin(ang)],
                        [0, 1, 0],
                        [-np.sin(ang), 0, np.cos(ang)]])

        for idx in range(0, 4):    # Post-synaptic neurons
            self.sd.neurons[idx]["rotation"] = R_x

        for idx in range(4, 8):   # Pre-synaptic neurons
            self.sd.neurons[idx]["rotation"] = R_y

        self.sd.detect(restart_detection_flag=True)

        synapse_voxel_loc = self.sd.hyper_voxel_synapses[:self.sd.hyper_voxel_synapse_ctr, 2:5]
        synapse_coords = synapse_voxel_loc * self.sd.voxel_size + self.sd.hyper_voxel_origo

        fig_path = os.path.join("networks", "network_testing_detect_cluster", "figures")

        if not os.path.exists(fig_path):
            os.mkdir(fig_path)

        if False:   # Set to True to include plot
            self.sd.plot_hyper_voxel(plot_neurons=True, fig_file_name="touch-detection-clusters-validation")

        # Check that each neuron pair
        with self.subTest(stage="cluster-size"):

            connection_list = [(4, 0, 5), (4, 1, 5), (4, 2, 5), (4, 3, 5),
                               (5, 0, 10), (5, 1, 10), (5, 2, 10), (5, 3, 10),
                               (6, 0, 10), (6, 1, 10), (6, 2, 10), (6, 3, 10),
                               (7, 0, 10), (7, 1, 10), (7, 2, 10), (7, 3, 10)]

            for pre_id, post_id, n_synapses in connection_list:
                self.assertTrue(self.check_neuron_pair_has_synapse(pre_id, post_id) == n_synapses)

        with self.subTest(stage="cluster-spread"):

            error_margin = np.sqrt(3*3**2)*2*1e-6  # Diagonal of either synapse voxel
            self.assertTrue(0 <= self.get_synapse_spread(4, 0) <= 25e-6 + error_margin)
            self.assertTrue(0 <= self.get_synapse_spread(4, 1) <= 25e-6 + error_margin)
            self.assertTrue(0 <= self.get_synapse_spread(4, 2) <= 25e-6 + error_margin)
            self.assertTrue(0 <= self.get_synapse_spread(4, 3) <= 25e-6 + error_margin)

        # Lägg till tester som kollar positionerna på synapserna
        # import pdb
        # pdb.set_trace()

    def check_neuron_pair_has_synapse(self, pre_neuron, post_neuron):

        connections = dict()

        for synapse_row in self.sd.hyper_voxel_synapses[0:self.sd.hyper_voxel_synapse_ctr, :]:

            loc = (synapse_row[0], synapse_row[1])
            if loc in connections:
                connections[loc] += 1
            else:
                connections[loc] = 1

        if (pre_neuron, post_neuron) in connections:
            # print(f"pre: {pre_neuron}, post: {post_neuron}, connections: {connections[(pre_neuron, post_neuron)]}")
            return connections[(pre_neuron, post_neuron)]
        else:
            # print(f"No connection between pre {pre_neuron} and post {post_neuron}")
            return False

    def get_synapse_spread(self, pre_neuron, post_neuron):

        coords = self.get_synapse_coordinates(pre_neuron, post_neuron)
        return np.max(pdist(coords))

    def get_synapse_coordinates(self, pre_neuron, post_neuron):

        coord_list = []

        for synapse_row in self.sd.hyper_voxel_synapses[0:self.sd.hyper_voxel_synapse_ctr, :]:
            if synapse_row[0] == pre_neuron and synapse_row[1] == post_neuron:
                coord_list.append(self.convert_to_coordinates(synapse_row[2:5]))

        return np.vstack(coord_list)

    def convert_to_coordinates(self, voxel_xyz):
        return np.array(voxel_xyz) * self.sd.voxel_size + self.sd.hyper_voxel_origo


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import h5py
import numpy as np
from lfpykit import CellGeometry, PointSourcePotential
import matplotlib.pyplot as plt

# You need to install lfpykit for this part

class LFP:

    def __init__(self, output_file):

        self.file = h5py.File(output_file, "r")

        self.extracellular_electrode_parameters = None

    def get_data(self, neuron_id):

        geometry = self.file["neurons"][str(neuron_id)]["geometry"][()].copy()
        membrane_current = self.file["neurons"][str(neuron_id)]["membrane.i_membrane_"][()]

        return geometry, membrane_current

    def set_electrode(self, sigma, x, y, z):

        self.extracellular_electrode_parameters = {
            "sigma": sigma,
            "x": np.array(x)*1e6,
            "y": np.array(y)*1e6,
            "z": np.array(z)*1e6   # micrometers for LFPy
        }

    def calculate_potential(self, neuron_id):

        geometry, membrane_current = self.get_data(neuron_id=neuron_id)

        cell_geometry = CellGeometry(x=geometry[:, [0, 1]]*1e6,
                                     y=geometry[:, [2, 3]]*1e6,
                                     z=geometry[:, [4, 5]]*1e6,
                                     d=geometry[:, [6, 7]]*1e6)

        forward_model = PointSourcePotential(cell_geometry, **self.extracellular_electrode_parameters)

        M = forward_model.get_transformation_matrix()
        V_e = M @ (membrane_current * 1e9)

        return V_e * 1e-3   # Convert to SI units

    def calculate_potential_network(self):

        V_e_list = []
        for neuron_id in self.file["neurons"].keys():
            V_e = self.calculate_potential(neuron_id=neuron_id)
            V_e_list.append(V_e)

        if len(V_e_list) > 1:
            return sum(V_e_list)
        else:
            return V_e

    def get_time(self):

        t = self.file["time"][()].copy()
        return t

    def plot_lfp(self, v_e, t):

        plt.plot(t, v_e.T)
        plt.xlabel("Time (s)")
        plt.ylabel("Voltage (V)")
        plt.ion()
        plt.show()


def cli():

    from argparse import ArgumentParser

    parser = ArgumentParser(description="Generate LFP")
    parser.add_argument("simulation_file")

    args = parser.parse_args()

    lfp = LFP(output_file=args.simulation_file)

    lfp.set_electrode(sigma=0.3, x=[0], y=[0], z=[0])
    v_e = lfp.calculate_potential_network()
    t = lfp.get_time()

    lfp.plot_lfp(v_e, t)

    input("Press a key to continue")

    import pdb
    pdb.set_trace()

if __name__ == "__main__":
    cli()


import os
import unittest

from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.detect.detect import SnuddaDetect
from snudda.utils.load import SnuddaLoad
from snudda.neurons.neuron_morphology_extended import NeuronMorphologyExtended
from snudda.place.place import SnuddaPlace
import numpy as np

from snudda.detect.prune import SnuddaPrune


class TestPrune(unittest.TestCase):
    
    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        self.network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_prune3")

        create_cube_mesh(file_name=os.path.join(self.network_path, "mesh", "simple_mesh.obj"),
                         centre_point=(0, 0, 0),
                         side_len=500e-6)

        config_file = os.path.join(self.network_path, "network-config.json")
        position_file = os.path.join(self.network_path, "network-neuron-positions.hdf5")
        save_file = os.path.join(self.network_path, "voxels", "network-putative-synapses.hdf5")

        sp = SnuddaPlace(config_file=config_file, d_view=None, verbose=True)

        sp.parse_config()
        sp.write_data(position_file)

        # We want to load in the ball and stick neuron that has 20 micrometer soma diameter, and axon (along y-axis),
        # and dendrite along (x-axis) out to 100 micrometer distance from centre of soma.

        from snudda.utils.reposition_neurons import RepositionNeurons
        repos = RepositionNeurons(position_file=position_file)

        # Reposition the neurons so that we know how many synapses and where they will be located before pruning
        # OBS, these positions and rotations are not written to the HDF5 file, they are only in memory!
        neuron_positions = np.array([[0, 20, 0],  # Postsynaptiska
                                     [0, 40, 0],
                                     [0, 60, 0],
                                     [0, 80, 0],
                                     [0, 100, 0],
                                     [0, 120, 0],
                                     [0, 140, 0],
                                     [0, 160, 0],
                                     [0, 180, 0],
                                     [0, 200, 0],
                                     [20, 0, 0],  # Presynaptiska
                                     [40, 0, 0],
                                     [60, 0, 0],
                                     [80, 0, 0],
                                     [100, 0, 0],
                                     [120, 0, 0],
                                     [140, 0, 0],
                                     [160, 0, 0],
                                     [180, 0, 0],
                                     [200, 0, 0],
                                     [70, 0, 500],  # For gap junction check
                                     [110, 0, 500],
                                     [150, 0, 500],
                                     [190, 0, 500],
                                     [0, 70, 500],
                                     [0, 110, 500],
                                     [0, 150, 500],
                                     [0, 190, 500],
                                     ]) * 1e-6

        # TODO: Add potential for gap junctions also by having 5 + 5 neurons in other grid
        
        ang = -np.pi / 2
        R_x = np.array([[1, 0, 0],
                        [0, np.cos(ang), -np.sin(ang)],
                        [0, np.sin(ang), np.cos(ang)]])
        
        ang = np.pi / 2
        R_y = np.array([[np.cos(ang), 0, np.sin(ang)],
                        [0, 1, 0],
                        [-np.sin(ang), 0, np.cos(ang)]])

        for idx in range(0, 10):  # Post synaptic neurons
            repos.place(neuron_id=idx, position=neuron_positions[idx, :], rotation=R_x)

        for idx in range(10, 20):  # Presynaptic neurons
            repos.place(neuron_id=idx, position=neuron_positions[idx, :], rotation=R_y)

        for idx in range(24, 28):  # GJ neurons
            repos.place(neuron_id=idx, position=neuron_positions[idx, :], rotation=R_x)

        ang = np.pi / 2
        R_z = np.array([[np.cos(ang), -np.sin(ang), 0],
                        [np.sin(ang), np.cos(ang), 0],
                        [0, 0, 1]])

        for idx in range(20, 24):  # GJ neurons
            repos.place(neuron_id=idx, position=neuron_positions[idx, :], rotation=np.matmul(R_z, R_x))

        self.sd = SnuddaDetect(config_file=config_file, position_file=position_file,
                               save_file=save_file, rc=None,
                               hyper_voxel_size=120, verbose=True)

        self.sd.detect(restart_detection_flag=True)

        if False:
            self.sd.process_hyper_voxel(1)
            self.sd.plot_hyper_voxel(plot_neurons=True,
                                     fig_file_name=os.path.join(self.network_path,
                                                                "test_prune_figure.png"))

            import pdb
            pdb.set_trace()

    def test_prune(self):

        pruned_output = os.path.join(self.network_path, "network-synapses.hdf5")

        with self.subTest(stage="No-pruning"):

            sp = SnuddaPrune(network_path=self.network_path, config_file=None, verbose=True, keep_files=True)  # Use default config file
            sp.prune()
            sp = []

            # Load the pruned data and check it

            sl = SnuddaLoad(pruned_output)
            # TODO: Call a plot function to plot entire network with synapses and all

            self.assertEqual(sl.data["num_synapses"], (20*8 + 10*2)*2)  # Update, now AMPA+GABA, hence *2 at end

            # This checks that all synapses are in order
            # The synapse sort order is destID, sourceID, synapsetype (channel model id).

            syn = sl.data["synapses"][:sl.data["num_synapses"], :]
            syn_order = (syn[:, 1] * len(self.sd.neurons) + syn[:, 0]) * 12 + syn[:, 6]  # The 12 is maxChannelModelID
            self.assertTrue((np.diff(syn_order) >= 0).all())

            # Note that channel model id is dynamically allocated, starting from 10 (GJ have ID 3)
            # Check that correct number of each type
            self.assertEqual(np.sum(sl.data["synapses"][:, 6] == 10), 20*8 + 10*2)
            self.assertEqual(np.sum(sl.data["synapses"][:, 6] == 11), 20*8 + 10*2)

            self.assertEqual(sl.data["num_gap_junctions"], 4*4*4)
            gj = sl.data["gap_junctions"][:sl.data["num_gap_junctions"], :2]
            gj_order = gj[:, 1] * len(self.sd.neurons) + gj[:, 0]
            self.assertTrue((np.diff(gj_order) >= 0).all())

        with self.subTest(stage="load-testing"):
            sl = SnuddaLoad(pruned_output, verbose=True)

            # Try and load a neuron
            n = sl.load_neuron(neuron_id=0)
            self.assertTrue(type(n) == NeuronMorphologyExtended)

            syn_ctr = 0
            for s in sl.synapse_iterator(chunk_size=50):
                syn_ctr += s.shape[0]
            self.assertEqual(syn_ctr, sl.data["num_synapses"])
        
            gj_ctr = 0
            for gj in sl.gap_junction_iterator(chunk_size=50):
                gj_ctr += gj.shape[0]
            self.assertEqual(gj_ctr, sl.data["num_gap_junctions"])

            syn, syn_coords = sl.find_synapses(pre_id=14)
            self.assertTrue((syn[:, 0] == 14).all())
            self.assertEqual(syn.shape[0], 40)

            syn, syn_coords = sl.find_synapses(post_id=3)
            self.assertTrue((syn[:, 1] == 3).all())
            self.assertEqual(syn.shape[0], 36)

            cell_id_perm = sl.get_neuron_id_of_type("ballanddoublestick", random_permute=True, num_neurons=28)
            cell_id = sl.get_neuron_id_of_type("ballanddoublestick", random_permute=False)

            self.assertEqual(len(cell_id_perm), 28)
            self.assertEqual(len(cell_id), 28)
            
            for cid in cell_id_perm:
                self.assertTrue(cid in cell_id)

        # It is important merge file has synapses sorted with dest_id, source_id as sort order since during pruning
        # we assume this to be able to quickly find all synapses on post synaptic cell.
        # TODO: Also include the ChannelModelID in sorting check
        with self.subTest("Checking-merge-file-sorted"):

            for mf in ["temp/synapses-for-neurons-0-to-28-MERGE-ME.hdf5",
                       "temp/gap_junctions-for-neurons-0-to-28-MERGE-ME.hdf5",
                       "network-synapses.hdf5"]:

                merge_file = os.path.join(self.network_path, mf)

                sl = SnuddaLoad(merge_file, verbose=True)
                if "synapses" in sl.data:
                    syn = sl.data["synapses"][:sl.data["num_synapses"], :2]
                    syn_order = syn[:, 1] * len(self.sd.neurons) + syn[:, 0]
                    self.assertTrue((np.diff(syn_order) >= 0).all())

                if "gap_junctions" in sl.data:
                    gj = sl.data["gap_junctions"][:sl.data["num_gap_junctions"], :2]
                    gj_order = gj[:, 1] * len(self.sd.neurons) + gj[:, 0]
                    self.assertTrue((np.diff(gj_order) >= 0).all())

        with self.subTest("synapse-f1"):
            # Test of f1
            testing_config_file = os.path.join(self.network_path, "network-config-test-1.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it

            sl = SnuddaLoad(pruned_output, verbose=True)
            # Setting f1=0.5 in config should remove 50% of GABA synapses, but does so randomly, for AMPA we used f1=0.9
            gaba_id = sl.data["connectivity_distributions"]["ballanddoublestick","ballanddoublestick"]["GABA"]["channel_model_id"]
            ampa_id = sl.data["connectivity_distributions"]["ballanddoublestick","ballanddoublestick"]["AMPA"]["channel_model_id"]

            n_gaba = np.sum(sl.data["synapses"][:, 6] == gaba_id)
            n_ampa = np.sum(sl.data["synapses"][:, 6] == ampa_id)

            # gaba: var(x) = n * p * (1-p), ie 180 * 0.5 * 0. 5 = 45 --> std(x) = 6.7, 2* std(x) = 13
            # ampa: 2*std(x) = 2*sqrt(180*0.9*0.1) = 8
            self.assertTrue((20*8 + 10*2)*0.5 - 13 < n_gaba < (20*8 + 10*2)*0.5 + 13)
            self.assertTrue((20*8 + 10*2)*0.9 - 8 < n_ampa < (20*8 + 10*2)*0.9 + 8)

        with self.subTest("synapse-softmax"):
            # Test of softmax
            testing_config_file = os.path.join(self.network_path, "network-config-test-2.json")  # Only GABA synapses in this config
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)
            # Softmax reduces number of synapses
            self.assertTrue(sl.data["num_synapses"] < 20*8 + 10*2)

        with self.subTest("synapse-mu2"):
            # Test of mu2
            testing_config_file = os.path.join(self.network_path, "network-config-test-3.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)
            # With mu2 having 2 synapses means 50% chance to keep them, having 1 will be likely to have it removed
            self.assertTrue(20*8*0.5 - 13 < sl.data["num_synapses"] < 20*8*0.5 + 13)

        with self.subTest("synapse-a3"):
            # Test of a3
            testing_config_file = os.path.join(self.network_path, "network-config-test-4.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)

            # a3=0.6 means 40% chance to remove all synapses between a pair
            self.assertTrue((20*8 + 10*2)*0.6 - 14 < sl.data["num_synapses"] < (20*8 + 10*2)*0.6 + 14)

        with self.subTest("synapse-distance-dependent-pruning"):
            # Testing distance dependent pruning
            testing_config_file = os.path.join(self.network_path, "network-config-test-5.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)

            # "1*(d >= 100e-6)" means we remove all synapses closer than 100 micrometers
            print(f"num synapses : {sl.data['num_synapses']}")

            self.assertEqual(sl.data["num_synapses"], 20*5)
            self.assertTrue((sl.data["synapses"][:, 8] >= 100).all())  # Column 8 -- distance to soma in micrometers

        # TODO: Need to do same test for Gap Junctions also -- but should be same results, since same codebase
        with self.subTest("gap-junction-f1"):
            # Test of f1
            testing_config_file = os.path.join(self.network_path, "network-config-test-6.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it

            sl = SnuddaLoad(pruned_output)
            # Setting f1=0.7 in config should remove 30% of gap junctions, but does so randomly
            self.assertTrue(64*0.7 - 10 < sl.data["num_gap_junctions"] < 64*0.7 + 10)

        with self.subTest("gap-junction-softmax"):
            # Test of softmax
            testing_config_file = os.path.join(self.network_path, "network-config-test-7.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)
            # Softmax reduces number of synapses
            self.assertTrue(sl.data["num_gap_junctions"] < 16*2 + 10)

        with self.subTest("gap-junction-mu2"):
            # Test of mu2
            testing_config_file = os.path.join(self.network_path, "network-config-test-8.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output)
            # With mu2 having 4 synapses means 50% chance to keep them, having 1 will be likely to have it removed
            self.assertTrue(64*0.5 - 10 < sl.data["num_gap_junctions"] < 64*0.5 + 10)

        with self.subTest("gap-junction-a3"):
            # Test of a3
            testing_config_file = os.path.join(self.network_path, "network-config-test-9.json")
            sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
            sp.prune()

            # Load the pruned data and check it
            sl = SnuddaLoad(pruned_output, verbose=True)

            # a3=0.7 means 30% chance to remove all synapses between a pair
            self.assertTrue(64*0.7 - 10 < sl.data["num_gap_junctions"] < 64*0.7 + 10)

        if False:  # Distance dependent pruning currently not implemented for gap junctions
            with self.subTest("gap-junction-distance-dependent-pruning"):
                # Testing distance dependent pruning
                testing_config_file = os.path.join(self.network_path, "network-config-test-10.json")
                sp = SnuddaPrune(network_path=self.network_path, config_file=testing_config_file, verbose=True, keep_files=True)  # Use default config file
                sp.prune()

                # Load the pruned data and check it
                sl = SnuddaLoad(pruned_output, verbose=True)

                # "1*(d <= 120e-6)" means we remove all synapses further away than 100 micrometers
                self.assertEqual(sl.data["num_gap_junctions"], 2*4*4)
                self.assertTrue((sl.data["gap_junctions"][:, 8] <= 120).all())  # Column 8 -- distance to soma in micrometers


if __name__ == '__main__':
    unittest.main()

# python3 -m unittest test_prune



import os
import numpy as np
import unittest

from snudda.neurons import NeuronMorphologyExtended
from snudda.place.bend_morphologies import BendMorphologies
import snudda.place.rotation

class TestBendMorphologies(unittest.TestCase):
    def test_something(self):

        base_path = os.path.dirname(__file__)

        morph_path = os.path.join(base_path, "validation", "striatum", "dspn",
                                  "str-dspn-e150602_c1_D1-mWT-0728MSN01-v20190508",
                                  "WT-0728MSN01-cor-rep-ax.swc")
        # mesh_path = os.path.join(base_path, "validation", "mesh", "Striatum-d-right.obj")
        mesh_path = os.path.join(base_path, "validation", "mesh", "bend_test_mesh.obj")

        nm = NeuronMorphologyExtended(swc_filename=morph_path)
        bm = BendMorphologies(mesh_path, rng=np.random.default_rng(12345))

        # pos = np.array([0.006, 0.004, 0.00205])
        pos = np.array([400e-6, 0, 0])


        rng = np.random.default_rng(1234)

        rot_mat = snudda.place.rotation.SnuddaRotate.rand_rotation_matrix(rng=rng)

        before = nm.clone(position=pos, rotation=rot_mat) # np.eye(3)
        after = nm.clone(position=pos, rotation=rot_mat) # np.eye(3)

        before_morph = before.get_morphology()
        old_rot_rep = bm.get_full_rotation_representation(morphology=before_morph)
        coords = bm.apply_rotation(morphology=before_morph, rotation_representation=old_rot_rep)

        # Verify that rotation representation works
        self.assertTrue((np.abs(before_morph.geometry[:, :3] - coords) < 1e-6).all())

        new_rot_rep, _ = bm.bend_morphology(after.get_morphology())

        new_coord = bm.apply_rotation(after.get_morphology(), new_rot_rep)
        after.get_morphology().geometry[:, :3] = new_coord

        change = np.sum(np.abs(before.get_morphology().geometry[:, :3] - after.get_morphology().geometry[:, :3]))
        print(f"Change = {change}")

        before_inside = bm.check_if_inside(before)
        after_inside = bm.check_if_inside(after)

        n_before = np.sum(before_inside)
        n_after = np.sum(after_inside)
        n_all = len(before_inside)

        # The bending is statistical, so we some parts of neuron might go a little outside
        self.assertTrue(n_after > n_before)

        self.assertTrue(n_after - n_before > 0.8 * (n_all - n_before))



if __name__ == '__main__':
    unittest.main()



import numpy as np

# fileName = "morphology/network/v7/LTS-20190212/LTS_papercorrected_savedcorrected_axon.swc"
file_name = "cellspecs/lts/LTS_Experiment-9862_20181211/Experiment-9862corrected-cor-rep.swc-withAxon"

# Load SWC file
with open(file_name, "r") as f:
    swcLines = f.readlines()

binWidth = 10e-6
dendArborHist = np.zeros((100,))
radius = np.arange(0, dendArborHist.shape[0]) * binWidth + binWidth / 2

swcVals = np.zeros(shape=(len(swcLines), 7))

comp_type = {1: "soma", 2: "axon", 3: "dend", 4: "apic"}

# Parse morphology
nComps = 0
for line in swcLines:
    if (line[0] != "#"):
        swcVals[nComps, :] = [float(s) for s in line.split()]
        nComps += 1

swcVals = swcVals[:nComps, :]

# Subtract 1 from ID and parentID, so we get easier indexing
swcVals[:, 0] -= 1
swcVals[:, 6] -= 1

swcVals[:, 2:6] *= 1e-6  # Convert to meter x,y,z, radie

# Plot neuron
import matplotlib.pyplot as plt

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')

# plt.plot(swcVals[:,2],swcVals[:,3],swcVals[:,4],'*')

assert swcVals[0, 1] == 1, "First row should be soma"

somaCoord = swcVals[0, 2:5]

for row in swcVals:
    if row[1] == 2:
        col = "r"
    elif (row[1] in [3, 4]):
        col = "b"
    else:
        continue

    parentCoords = swcVals[int(row[6]), 2:5] - somaCoord
    coords = row[2:5] - somaCoord
    centre = (coords + parentCoords) / 2.0

    plt.plot([coords[0], parentCoords[0]],
             [coords[1], parentCoords[1]],
             [coords[2], parentCoords[2]], col + '-')

########################################################################

nPoints = 100000
rng = [-100e-6, 1500e-6, -200e-6, 200e-6, -200e-6, 200e-6]

x = rng[0] + (rng[1] - rng[0]) * np.random.rand(nPoints, 1)
y = rng[2] + (rng[3] - rng[2]) * np.random.rand(nPoints, 1)
z = rng[4] + (rng[5] - rng[4]) * np.random.rand(nPoints, 1)

x1 = 200e-6  # Inner bump
xw1 = 100e-6  # 100e-6
x2 = 300e-6  # Tube
xw2 = 300e-6
x3 = 700e-6  # Outer bump
xw3 = 100e-6

Pkeep = 0.25 * np.exp(-(((x - x1) / xw1) ** 2 + ((y - 0) / 50e-6) ** 2 + ((z - 0) / 30e-6) ** 2)) \
        + 1 * np.exp(-(((x - x2) / xw2) ** 2 + ((y - 0) / 15e-6) ** 2 + ((z - 0) / 10e-6) ** 2)) \
        + 1 * np.exp(-(((x - x3) / xw3) ** 2 + ((y - 0) / 15e-6) ** 2 + ((z - 0) / 15e-6) ** 2))

Pkeep = Pkeep / np.max(Pkeep)

idx = np.where(Pkeep > np.random.rand(nPoints, 1))[0]
xkeep = x[idx]
ykeep = y[idx]
zkeep = z[idx]

dall = np.sqrt(xkeep ** 2 + ykeep ** 2 + zkeep ** 2)

# import matplotlib.pyplot as plt
# from mpl_toolkits.mplot3d import Axes3D

# fig = plt.figure()
# ax = fig.add_subplot(111, projection='3d')

ax.scatter(0, 0, 0, '*', c="g")
ax.scatter(xkeep, ykeep, zkeep, '*', c='k')
ax.set_aspect("equal")

# ax.set_xlim([rng[0],rng[1]])

plt.ion()
plt.show()

plt.savefig("figures/LTS-axon-density-tuning-3D.png")

# Read the data from Tepper
tepperData = "DATA/Tepper2011extracted.csv"
teppersholldata = np.genfromtxt(tepperData, delimiter=",")

xsholl = teppersholldata[1:, 0]
axonsholl = teppersholldata[1:, 1]
dendsholl = teppersholldata[1:, 2]

fig, ax = plt.subplots()
ax.hist(dall)
# plt.scatter(xsholl*1e-6, axonsholl,c="r")
ax2 = ax.twinx()

ax2.plot(xsholl * 1e-6, axonsholl, "*", c="r")
ax.set_xlabel("Distance from soma")
ax.set_ylabel("Axon count")
ax2.set_ylabel("Axon sholl count")

plt.ion()
plt.show()

plt.savefig("figures/LTS-axon-density-tuning-histogram.png")

import pdb

pdb.set_trace()



import numpy as np

import json
from copy import deepcopy

from snudda import SnuddaLoad
from snudda.utils.swap_to_degenerated_morphologies import SwapToDegeneratedMorphologies

# TODO: We need to reapply the pruning rule to all neurons...


class SwapToDegeneratedMorphologiesExtended(SwapToDegeneratedMorphologies):

    def __init__(self, original_network_file, updated_network_file, output_network_file,
                 original_snudda_data_dir, updated_snudda_data_dir,
                 original_input_file=None, updated_input_file=None, output_input_file=None,
                 post_degen_pruning=True,
                 filter_axon=True,
                 random_seed=None
                 ):

        """
            Args:
                original_network_file : Original network-synapses.hdf5 (e.g. network with WT morphologies)
                updated_network_file : network-synapses.hdf5 network with updated morphologies (e.g. Parkinson network)
                output_network_file : Created network-synapses.hdf5 network by this code
                original_snudda_data_dir : Path to original data (e.g. WT SNUDDA_DATA)
                updated_snudda_data_dir : Path to updated data (e.g. Parkinson SNUDDA_DATA)
                original_input_file : Input to original network (e.g. WT network input)
                updated_input_file : Input to the updated_network_file network
                output_input_file: Input file generated by this code (great name!)
                post_degen_pruning: Should the neurons be pruned again after degeneration (and growth)
                filter_axon : Degeneration of axon leads to removal of synapses
                random_seed: Random seed used for post pruning

        """

        super().__init__(original_network_file=original_network_file,  # PD0
                         new_network_file=output_network_file,         # degenerated PD2 without FS axon
                         original_snudda_data_dir=original_snudda_data_dir,
                         new_snudda_data_dir=updated_snudda_data_dir,
                         original_input_file=original_input_file,
                         new_input_file=output_input_file, filter_axon=filter_axon)

        self.updated_network_file = updated_network_file  # PD2 fresh touch detection, generated before
        self.updated_network_loader = SnuddaLoad(self.updated_network_file, load_synapses=False)
        self.updated_hdf5 = self.updated_network_loader.hdf5_file
        self.updated_data = self.updated_network_loader.data
        self.random_seed = random_seed

        self.old_kd_tree_cache = dict()
        self.old_neuron_cache_id = dict()

        self.check_same_network()
        self.rng = np.random.default_rng(seed=random_seed)

        self.post_degen_pruning = post_degen_pruning

    def close(self):

        super().close()

        if self.updated_hdf5:
            self.updated_hdf5.close()

    def check_same_network(self):

        assert len(self.original_network_loader.data["neurons"]) == len(self.updated_network_loader.data["neurons"]),\
            f"Original and updated network does not have the same number of neurons!"

        for orig_neuron, updated_neuron in zip(self.original_network_loader.data["neurons"],
                                               self.updated_network_loader.data["neurons"]):

            try:
                assert orig_neuron["neuron_id"] == updated_neuron["neuron_id"], f"Internal error, neuron ID mismatch"

                assert orig_neuron["name"] == updated_neuron["name"], \
                    (f"Name mismatch for neuron {orig_neuron['neuron_id']}: {orig_neuron['name']} {updated_neuron['name']}"
                     f"\nDid you use the same random seed when calling init to generate the networks? "
                     f"Also, did you use the same number of workers for both? The neurons are clustered differently "
                     f"depending on number of workers, to help improve speed.")

                assert (orig_neuron["position"] == updated_neuron["position"]).all(), \
                    (f"Position mismatch for neuron {orig_neuron['neuron_id']}: "
                     f"{orig_neuron['position']} {updated_neuron['position']}"
                     f"\nDid you use the same random seed when calling init to generate the networks?")

                assert (orig_neuron["rotation"] == updated_neuron["rotation"]).all(), \
                    (f"Position mismatch for neuron {orig_neuron['neuron_id']}: " 
                     f"{orig_neuron['rotation']} {updated_neuron['rotation']}"
                     f"\nDid you use the same random seed when calling init to generate the networks?")
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

    def get_degeneration_recovery_lookups(self, network_config):

        type_lookup = self.updated_network_loader.get_neuron_types()
        degeneration_recovery = dict()

        for region_name, region_data in network_config["regions"].items():

            for conf_key, conf_info in region_data["connectivity"].items():
                pre_type, post_type = conf_key.split(",")

                for con_type, con_info in conf_info.items():
                    channel_model_id = self.updated_network_loader.data["connectivity_distributions"][pre_type,post_type][con_type]["channel_model_id"]

                    if "degeneration_recovery" in con_info["pruning"]:
                        degeneration_recovery[pre_type, post_type, channel_model_id] = con_info["pruning"]["degeneration_recovery"]

        return degeneration_recovery, type_lookup

    def get_additional_synapses(self, network_config, rng, synapse_distance_treshold=None):

        # Calculate coordinate remapping for updated synapses

        voxel_size = self.updated_network_loader.data["voxel_size"]
        assert voxel_size == self.original_network_loader.data["voxel_size"], f"Voxel size mismatch between networks"

        if synapse_distance_treshold is None:
            synapse_distance_treshold = 1.2*(3 * voxel_size ** 2) ** 0.5  # 5.2e-6, maximal mismatch due to moving origos, actually 5.4 micrometers

        orig_sim_origo = self.original_network_loader.data["simulation_origo"]
        updated_sim_origo = self.updated_network_loader.data["simulation_origo"]

        origo_diff = updated_sim_origo - orig_sim_origo
        voxel_transform = np.round(origo_diff / voxel_size)

        # Will hold indexes into synapse matrix, so we can do all synapses belonging to
        # one morphology all at once
        pre_neuron_synapses = dict()
        post_neuron_synapses = dict()

        for nid in self.updated_network_loader.get_neuron_id():
            pre_neuron_synapses[nid] = []
            post_neuron_synapses[nid] = []

        # All the synapses from the fresh PD2 touch detection
        # -- This copy should be conditional on whether it is hdf5 read data, or in memory
        synapse_matrix = self.updated_network_loader.data["synapses"][()].copy()

        keep_mask = np.zeros((synapse_matrix.shape[0],), dtype=bool)

        for idx, synapse_row in enumerate(synapse_matrix):
            pre_neuron_synapses[synapse_row[0]].append(idx)
            post_neuron_synapses[synapse_row[1]].append(idx)

        for nid in self.updated_network_loader.get_neuron_id():
            if nid % 100 == 0:
                print(f"Processing neuron {nid}")

            pre_idx = np.array(pre_neuron_synapses[nid], dtype=int)
            post_idx = np.array(post_neuron_synapses[nid], dtype=int)

            pre_coords = synapse_matrix[pre_idx, 2:5] * voxel_size + updated_sim_origo
            post_coords = synapse_matrix[post_idx, 2:5] * voxel_size + updated_sim_origo

            morph = self.get_morphology(neuron_id=nid, hdf5=self.old_hdf5,
                                        neuron_cache_id=self.old_neuron_cache_id,
                                        snudda_data=self.original_snudda_data_dir)

            dend_kd_tree = self.get_kd_tree(morph, "dend", kd_tree_cache=self.old_kd_tree_cache)
            synapse_dend_dist, _ = dend_kd_tree.query(post_coords)
            soma = morph.morphology_data["neuron"].sections[1][0]
            syn_mask = np.logical_and(synapse_dend_dist > synapse_distance_treshold,
                                      np.linalg.norm(post_coords - soma.position, axis=1)
                                      > soma.radie + synapse_distance_treshold)
            keep_mask[post_idx[np.where(syn_mask)[0]]] = True

            if self.updated_network_loader.data["neurons"][nid]["axon_density"] is None:
                try:
                    axon_kd_tree = self.get_kd_tree(morph, "axon", kd_tree_cache=self.old_kd_tree_cache)
                    synapse_axon_dist, _ = axon_kd_tree.query(pre_coords)
                    keep_mask[pre_idx[np.where(synapse_axon_dist > synapse_distance_treshold)[0]]] = True

                    # if nid == 100:
                    #     print("Tell me why .. axon")
                    #     import pdb
                    #     pdb.set_trace()

                except:
                    import traceback
                    print(traceback.format_exc())
                    import pdb
                    pdb.set_trace()
            else:
                print(f"No axon for neuron {morph.name} ({nid})")

        print("Synapse degeneration recovery...")
        # If degenerationRecovery is set in the pruning rules, we will add that fraction of the synapses on the
        # overlapping morphology part back to the network synapses -- and take care not to place synapses
        # exactly where old synapses are already placed

        # Get information about synapse degeneration recovery (ie how large a fraction of the synapses
        # that are in the overlaping part of the dendrites should be kept to recover a little...)
        degeneration_recovery, type_lookup = self.get_degeneration_recovery_lookups(network_config=network_config)

        if len(degeneration_recovery) > 0:
            p_recovery = np.zeros((synapse_matrix.shape[0],), dtype=float)

            for idx, syn_row in enumerate(synapse_matrix):
                pre_type = type_lookup[syn_row[0]]
                post_type = type_lookup[syn_row[1]]
                channel_model_id = syn_row[6]

                if (pre_type, post_type, channel_model_id) in degeneration_recovery:
                    p_recovery[idx] = degeneration_recovery[pre_type, post_type, channel_model_id]

            recovery_mask = p_recovery > rng.random(p_recovery.shape)
            keep_mask = np.logical_or(keep_mask, recovery_mask)

        # Transform coordinates to new simulation origo
        added_synapses = synapse_matrix[keep_mask, :].copy()
        added_synapses[:, 2:5] = added_synapses[:, 2:5] + voxel_transform

        return added_synapses

    def sort_synapses(self, synapses):

        # Sort order: columns 1 (dest), 0 (src), 6 (synapse type)
        sort_idx = np.lexsort(synapses[:, [6, 0, 1]].transpose())
        return synapses[sort_idx, :].copy()

    def filter_synapses(self, filter_axon=False, post_degen_pruning=None):

        # This replaces the original filter_synapses, so that we can also add in the
        # new synapses due to growing axons or dendrites

        if post_degen_pruning is None:
            post_degen_pruning = self.post_degen_pruning

        config = deepcopy(self.updated_network_loader.data["config"])

        # This needs to be made bigger!
        num_rows = self.old_hdf5["network/synapses"].shape[0] + self.updated_hdf5["network/synapses"].shape[0]
        num_cols = self.old_hdf5["network/synapses"].shape[1]
        new_synapses = np.zeros((num_rows, num_cols), dtype=self.old_hdf5["network/synapses"].dtype)
        syn_ctr = 0

        # Keep synapses in original network that are still within the new morphologies
        for synapses in self.synapse_iterator():

            # if synapses[0, 0] == 4 and synapses[0, 1] == 100:
            #     print("tell me why...")
            #     import pdb
            #     pdb.set_trace()

            new_syn = self.filter_synapses_helper(synapses, filter_axon=filter_axon)
            new_synapses[syn_ctr:syn_ctr + new_syn.shape[0]] = new_syn
            syn_ctr += new_syn.shape[0]

        # Here add the new synapses from growing axons and dendrites
        additional_synapses = self.get_additional_synapses(network_config=config, rng=self.rng)
        new_synapses[syn_ctr:syn_ctr+additional_synapses.shape[0], :] = additional_synapses
        syn_ctr += additional_synapses.shape[0]

        # Resort the new synapse matrix
        sorted_synapses = self.sort_synapses(new_synapses[:syn_ctr, :])

        # Copy the synapses into memory
        old_synapse_iterator = self.synapse_iterator(synapses=self.old_hdf5["network/synapses"][()].copy())

        if post_degen_pruning:
            pruned_synapses = self.post_degeneration_pruning(synapses=sorted_synapses,
                                                             old_synapse_iterator=old_synapse_iterator,
                                                             network_config=config,
                                                             rng=self.rng)
        else:
            pruned_synapses = sorted_synapses

        sorted_synapses = None

        # TODO: We should check that the addition of "degeneratio recovery synapses" are not duplicates
        #       of the PD degeneration synapses (this might happen in 8% of the cases...)

        self.new_hdf5["network"].create_dataset("synapses", data=pruned_synapses, compression="lzf")

        num_synapses = np.zeros((1,), dtype=np.uint64)
        self.new_hdf5["network"].create_dataset("num_synapses", data=pruned_synapses.shape[0], dtype=np.uint64)

        print(f"Keeping {self.new_hdf5['network/num_synapses'][()]} "
              f"out of {self.old_hdf5['network/num_synapses'][()]} synapses "
              f"({self.new_hdf5['network/num_synapses'][()] / self.old_hdf5['network/num_synapses'][()]*100:.3f} %)")

    def post_degeneration_pruning(self, synapses, old_synapse_iterator, network_config, rng):

        """ Args:
            synapses: Synapse matrix to prune
            old_synapse_iterator: Iterator over the original synapses
            network_config: The config as a dictionary
            rng: Numpy random generator
        """

        # TODO: We here assume that mu2 is the same for WT and degenerate network, if it is NOT the same
        #       this code needs to be modified such that mu2_original and mu2_degenerate are used

        # This prunes the network synapses after the generation/growth phase
        print("Running post degeneration pruning of synapses", flush=True)

        keep_synapse_flag = np.ones((synapses.shape[0]), dtype=bool)

        type_lookup = self.updated_network_loader.get_neuron_types()
        mu2_lookup = dict()
        n_neurons = self.updated_network_loader.data["num_neurons"]

        for region_name, region_data in network_config["regions"].items():

            for conf_key, conf_info in region_data["connectivity"].items():
                pre_type, post_type = conf_key.split(",")

                if pre_type not in mu2_lookup:
                    mu2_lookup[pre_type] = dict()

                if post_type not in mu2_lookup[pre_type]:
                    mu2_lookup[pre_type][post_type] = dict()

                for con_type, con_info in conf_info.items():

                    channel_model_id = self.updated_network_loader.data["connectivity_distributions"][pre_type,post_type][con_type]["channel_model_id"]

                    # OBS, this does not take into account population units. Neurons belonging to different population unit
                    # can have a different mu2, but we have never used that option.
                    mu2_lookup[pre_type][post_type][channel_model_id] = con_info["pruning"]["mu2"]

        synapse_ctr = 0

        old_synapse_set = next(old_synapse_iterator, None)

        p_mu_overflow = []

        for synapse_set in self.synapse_iterator(synapses=synapses):

            pre_id = synapse_set[0, 0]
            post_id = synapse_set[0, 1]
            channel_model_id = synapse_set[0, 6]

            try:
                assert (synapse_set[:, 0] == pre_id).all()
                assert (synapse_set[:, 1] == post_id).all()
                assert (synapse_set[:, 6] == channel_model_id).all(), \
                    f"Code is written with assumption that all synapses between a pair of neurons are of the same type"
            except:
                print(f"Synapse set between: {self.original_network_loader.data['neurons'][pre_id]} and {self.original_network_loader.data['neurons'][post_id]}")
                print(f"{synapse_set}")
                print(f"Possible error: You used GapJunction instead of gap_junction for FS-FS connectivity in network json file")
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

            if old_synapse_set is not None:
                old_pre_id = old_synapse_set[0, 0]
                old_post_id = old_synapse_set[0, 1]
                old_channel_mod_id = old_synapse_set[0, 6]
            else:
                old_pre_id = -1
                old_post_id = -1
                old_channel_mod_id = None

            while old_synapse_set is not None and old_post_id * n_neurons + old_pre_id < post_id * n_neurons + pre_id:
                old_synapse_set = next(old_synapse_iterator, None)

                if old_synapse_set is not None:
                    old_pre_id = old_synapse_set[0, 0]
                    old_post_id = old_synapse_set[0, 1]
                    old_channel_mod_id = old_synapse_set[0, 6]
                else:
                    old_pre_id = -1
                    old_post_id = -1
                    old_channel_mod_id = None

            try:
                mu2 = mu2_lookup[type_lookup[pre_id]][type_lookup[post_id]][channel_model_id]
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

            n_syn = synapse_set.shape[0]

            if mu2 is None:
                p_mu = 1

            else:

                if old_post_id * n_neurons + old_pre_id == post_id * n_neurons + pre_id:
                    old_n_syn = old_synapse_set.shape[0]

                    # This should be mu2 original
                    old_p_mu = 1.0 / (1.0 + np.exp(-8.0 / mu2 * (old_n_syn - mu2)))

                    assert old_channel_mod_id is None or old_channel_mod_id == channel_model_id, \
                        (f"post_degeneration_pruning: Internal error, we have assumed only one type of synapses between "
                         f"pairs of neurons in this degenerated code.\n"
                         f"channel_mod_id = {channel_model_id}, old_channel_mod_id = {old_channel_mod_id}\n"
                         f"synapse_set = {synapse_set}\n"
                         f"old_synapse_set = {old_synapse_set}")

                else:
                    old_p_mu = 1

                # This should use mu2_degenerated (which is does, but mu2_old above
                # should use mu2_original which id does not)
                # TODO: Fix it or add assert that they are the same, right now it will
                #  silently do it wrong if different...

                # We need to compensate for the old p_mu, i.e. p_real = p_mu_new / p_mu_old
                p_mu = 1.0 / (1.0 + np.exp(-8.0 / mu2 * (n_syn - mu2)))
                # print(f"p_mu = {p_mu} ({p_mu / old_p_mu}) -- {old_p_mu}")
                p_mu /= old_p_mu  # Correction factor for previous mu2 pruning

            keep_synapse_flag[synapse_ctr:synapse_ctr + n_syn] = p_mu >= rng.random()
            synapse_ctr += n_syn

            if p_mu > 1:
                p_mu_overflow.append((p_mu, n_syn))

        # Check how many synapses we are expected to be short...
        n_short = 0
        for pm, ns in p_mu_overflow:
            n_short += (pm - 1) * ns

        print(f"Unable to compensate for {n_short:.1f} degenerated synapses.")

        print(f"Post pruning. Keeping {np.sum(keep_synapse_flag)}/{len(keep_synapse_flag)} "
              f"({np.sum(keep_synapse_flag)/len(keep_synapse_flag)*100:.3f}%)")

        return synapses[keep_synapse_flag, :]


    # TODO: Profile the code to see what the bottleneck is...

    # TODO: Update filter_gap_junctions to also handle growth on dendrites

    # TODO: Also handle the new inputs that might arrive on growing morphologies...



import json
import os
import timeit
from collections import OrderedDict


class BenchmarkLogging:
    """ Saves benchmark logging when running snudda from command line.

    Trivial example below:

    import time
    bl = BenchmarkLogging("/home/hjorth/HBP/Snudda/snudda")
    bl.start_timer("test")
    time.sleep(1)
    bl.start_timer("test2")
    time.sleep(2)
    bl.stop_timer("test2")
    bl.stop_timer("test")
    bl.write_log()

    """

    def __init__(self, network_path, parallel_flag=False, log_file=None, running_neuron=False,
                 ipython_profile=None):

        """
        Constructor.

        Args:
            network_path (str): Path to network
            parallel_flag (bool): Running in parallel, should we determine number of workers?
            log_file (str) : Log file to save text to
            running_neuron (bool) : Are we running NEURON? (Sets method for determining number of workers)
        """

        if log_file:
            self.log_file = log_file
        else:
            self.log_file = os.path.join(network_path, "benchmark_log.json")

        self.network_name = self.get_network_name(network_path)

        self.start_time = dict()
        self.end_time = dict()
        self.pc = None  # Used if running neuron

        if parallel_flag or running_neuron:
            self.num_workers = self.get_number_of_workers(running_neuron=running_neuron,
                                                          ipython_profile=ipython_profile)
        else:
            self.num_workers = 1

    def get_number_of_workers(self, running_neuron, ipython_profile=None):

        """
        Returns number of workers.

        Args:
            running_neuron (bool) : Are we running NEURON? Used when determining number of workers).

        """

        if running_neuron:
            # We are running neuron, different way to detect number of workers
            from mpi4py import MPI   # Import mpi before NEURON for parallel
            from neuron import h
            self.pc = h.ParallelContext()
            return self.pc.nhost()

            # Is there a simpler way to get the number of workers?

        if ipython_profile is None:
            ipython_profile = os.getenv('IPYTHON_PROFILE')

        if not ipython_profile:
            ipython_profile = "default"

        ipython_dir = os.getenv('IPYTHONDIR')
        if not ipython_dir:
            ipython_dir = os.path.join(os.path.abspath(os.getcwd()), ".ipython")

        import ipyparallel
        u_file = os.path.join(ipython_dir, f"profile_{ipython_profile}", "security", "ipcontroller-client.json")

        if os.path.isfile(u_file):
            print(f"Benchmark reading ipyparallel config file: {u_file}")
            try:
                rc = ipyparallel.Client(url_file=u_file, profile=ipython_profile, timeout=120, debug=False)
                d_view = rc.direct_view(targets='all')  # rc[:] # Direct view into clients

                return len(d_view) + 1  # We also include the master node
            except:
                return -1
        else:
            return 1

    @staticmethod
    def get_network_name(network_path):

        """ Returns network name based on network_path. """

        if os.path.basename(network_path):
            network_name = os.path.basename(network_path)
        else:
            network_name = os.path.basename(os.path.dirname(network_path))

        return network_name

    def start_timer(self, item_name):

        """ Start benchmark timer for item_name. """

        self.start_time[item_name] = timeit.default_timer()

    def stop_timer(self, item_name):
        """ Stops benchmark timer for item_name. """
        self.end_time[item_name] = timeit.default_timer()

    def write_log(self):

        """ Writes to log. """

        if self.pc and self.pc.id() != 0:
            # If this is true we are running NEURON, and with id != 0 we are not master node, just return
            return

        if os.path.exists(self.log_file):
            try:
                with open(self.log_file, "r") as fr:
                    data = json.load(fr, object_pairs_hook=OrderedDict)
            except:
                print(f"Error loading {self.log_file}, creating new benchmark log.")
                # Start with fresh data, this will overwrite old data
                data = OrderedDict()

        else:
            data = OrderedDict()

        if self.network_name not in data:
            data[self.network_name] = OrderedDict()

        del_keys = []

        for item_name in self.end_time.keys():
            assert item_name in self.start_time, f"Benchmark logging was not started for {item_name}"
            duration = self.end_time[item_name] - self.start_time[item_name]

            if item_name in data[self.network_name]:
                data[self.network_name][item_name].append([duration, self.num_workers])
            else:
                data[self.network_name][item_name] = [[duration, self.num_workers]]

            del_keys.append(item_name)

        for key in del_keys:
            # Remove the old start and end times
            del self.start_time[key]
            del self.end_time[key]

        with open(self.log_file, "w") as fw:
            json.dump(data, fw, indent=4)



import os
import numpy as np
import h5py

import unittest

from snudda import SnuddaDetect, SnuddaPrune
from snudda.plotting import PlotNetwork
from snudda.utils.swap_to_degenerated_morphologies_extended import SwapToDegeneratedMorphologiesExtended


class MyTestCase(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        network_A = os.path.join("networks", "network_testing_degeneration", "A")
        network_B = os.path.join("networks", "network_testing_degeneration", "B")
        network_C = os.path.join("networks", "network_testing_degeneration", "C")

        self.network_A = network_A
        self.network_C = network_C

        from snudda.place import SnuddaPlace

        # Setup original network
        # https://www.americanscientist.org/article/the-quest-for-randomness
        sp = SnuddaPlace(network_path=network_A, random_seed=9)
        sp.place()

        network_A_pos_file = os.path.join(network_A, "network-neuron-positions.hdf5")
        self.set_neuron_positions(network_file=network_A_pos_file)

        sd = SnuddaDetect(network_path=network_A, random_seed=11)
        sd.detect()

        spr = SnuddaPrune(network_path=network_A, random_seed=13)
        spr.prune()

        # Setup degenerated network
        sp2 = SnuddaPlace(network_path=network_B, random_seed=9)  # same as original
        sp2.place()

        network_B_pos_file = os.path.join(network_B, "network-neuron-positions.hdf5")
        self.set_neuron_positions(network_file=network_B_pos_file)

        sd2 = SnuddaDetect(network_path=network_B, random_seed=11)
        sd2.detect()

        spr2 = SnuddaPrune(network_path=network_B, random_seed=13)
        spr2.prune()

        network_A_file = os.path.join(network_A, "network-synapses.hdf5")
        network_B_file = os.path.join(network_B, "network-synapses.hdf5")
        network_C_file = os.path.join(network_C, "network-synapses.hdf5")

        # These paths should normally be to the SNUDDA_DATA directory, and not directly to the neuron folders
        # but because we only have one neuron type in our model in this test, this will work... for now.
        original_data_dir = os.path.join("validation", "ballanddoublestick_original")
        updated_data_dir = os.path.join("validation", "ballanddoublestick_degenerated")

        swap = SwapToDegeneratedMorphologiesExtended(original_network_file=network_A_file,
                                                     updated_network_file=network_B_file,
                                                     output_network_file=network_C_file,
                                                     original_snudda_data_dir=original_data_dir,
                                                     updated_snudda_data_dir=updated_data_dir,
                                                     random_seed=15)

        swap.write_new_network_file()

        # Create config files for the network using ballanddoublestick, and ballanddoublestick_degenerated
        # (Here axons are elongated from 220 to 300 micrometers, and dendrites are shrunk from 200 to 150 micrometers)

        # Create placement files for the WT and degenerated networks, important that they both have the same neurons.

        # Do touch detection both networks

        # Do swap_to_degenerated_morphologies_extended

        pn = PlotNetwork(network_A)
        pn.plot(plot_axon=True, plot_dendrite=True, plot_synapses=True, fig_name="original-network.png")
        pn.close()

        pn2 = PlotNetwork(network_B)
        pn2.plot(plot_axon=True, plot_dendrite=True, plot_synapses=True, fig_name="degenerated-network.png")
        pn2.close()

        pn3 = PlotNetwork(network_C)
        pn3.plot(plot_axon=True, plot_dendrite=True, plot_synapses=True, fig_name="degenerated-network.png")
        pn3.close()

    def set_neuron_positions(self, network_file):

        with h5py.File(network_file, "r+") as hdf5_file:

            neuron_positions = np.array([[0, 20, 0],  # Postsynaptiska
                                         [0, 40, 0],
                                         [0, 60, 0],
                                         [0, 80, 0],
                                         [0, 100, 0],
                                         [0, 120, 0],
                                         [0, 140, 0],
                                         [0, 160, 0],
                                         [0, 180, 0],
                                         [0, 200, 0],
                                         [20, 0, 0],  # Presynaptiska
                                         [40, 0, 0],
                                         [60, 0, 0],
                                         [80, 0, 0],
                                         [100, 0, 0],
                                         [120, 0, 0],
                                         [140, 0, 0],
                                         [160, 0, 0],
                                         [180, 0, 0],
                                         [200, 0, 0],
                                         [100, 100, -240],  # To get a gap junction
                                         ]) * 1e-6

            hdf5_file["network/neurons/position"][:, :] = neuron_positions

            ang = -np.pi / 2
            R_x = np.array([[1, 0, 0],
                            [0, np.cos(ang), -np.sin(ang)],
                            [0, np.sin(ang), np.cos(ang)]])

            ang = np.pi / 2
            R_y = np.array([[np.cos(ang), 0, np.sin(ang)],
                            [0, 1, 0],
                            [-np.sin(ang), 0, np.cos(ang)]])

            ang = -np.pi / 2
            R_gj = np.array([[np.cos(ang), 0, np.sin(ang)],
                             [0, 1, 0],
                             [-np.sin(ang), 0, np.cos(ang)]])

            for idx in range(0, 10):  # Post synaptic neurons
                hdf5_file["network/neurons/rotation"][idx, :] = R_x.flatten()

            for idx in range(10, 20):  # Presynaptic neurons
                hdf5_file["network/neurons/rotation"][idx, :] = R_y.flatten()

                hdf5_file["network/neurons/rotation"][20, :] = R_gj.flatten()

            hdf5_file.close()

    def test_something(self):

        # Load the networks
        from snudda.utils import SnuddaLoad
        orig_load = SnuddaLoad(self.network_A)
        degen_load = SnuddaLoad(self.network_C)

        tmp = [(x["neuron_id"], x["morphology_key"]) for x in orig_load.data["neurons"]]
        print(f"Morphologies: {tmp}")

        #import pdb
        #pdb.set_trace()

        # self.assertEqual(orig_load.data["num_synapses"], 165)

        # Verify that it should be 99 synapses -- now it is just a regression test
        # Old version gave 99, new gives 165 --- CHECK WHY!

        # TODO: CHECK WHY NOT 99 SYNAPSES NOW
        # self.assertEqual(degen_load.data["num_synapses"], 155)  # -- Ilaria, we need to check what the true value should be?


if __name__ == '__main__':
    unittest.main()



import json
import argparse

def check_missing(dict_a, dict_b, file_b):
  for k in dict_a.keys():
    if k not in dict_b:
      print(f"Parameter key {k} missing in {file_b}")
    else:
      for k2 in dict_a[k].keys():
        if k2 not in dict_b[k]:
          print(f"Morphology key {k2} missing for parameter set {k} in {file_b}")
      
def cli():
  parser = argparse.ArgumentParser()
  parser.add_argument("file_a")
  parser.add_argument("file_b")

  args = parser.parse_args()

  with open(args.file_a, "r") as f:
    dict_a = json.load(f)

  with open(args.file_b, "r") as f:
    dict_b = json.load(f)

  print(f"Comparing {args.file_a} and {args.file_b}")
    
  check_missing(dict_a, dict_b, args.file_b)
  check_missing(dict_b, dict_a, args.file_a)

  
cli()



# 1. Load master_config into master dictionary
# 2. Identify slave configs (configs referred in master_config), and load them into master dictionary
# 3. Enrich master dictionary, e.g with additional neuron information
# 4. Additional modifications?
# 5. Export network-config.json

import os
import json
from copy import deepcopy
from collections import ChainMap
from snudda.utils import snudda_parse_path
from snudda.utils import NumpyEncoder


class ConfigParser:

    def __init__(self, config_file=None, snudda_data=None, parse=True):

        self.config_file = config_file
        self.network_info = dict()  # master dictionary
        self.snudda_data = snudda_data
        self.config_data = {}
        self.rng = None

        # By default, JSON files which are included in the config file are parsed and inserted in place
        # if you instead want to keep the reference to the file, then add it below:
        self.exclude_parse_keys = ["parameter_file", "projection_file", "rotation_file", "rotation_field_file",
                                   "reaction_diffusion", "modulation", "extracellular_config"]
        self.exclude_parse_values = ["parameters.json", "mechanisms.json", "modulation.json", "meta.json"]

        if config_file is not None and parse:
            self.load()

    def load(self):
        print(f"Loading {self.config_file}")

        if os.path.isfile(self.config_file):
            with open(self.config_file) as f:
                self.config_data = json.load(f)
        else:
            raise ValueError(f"File not found: {self.config_file}")

        if self.snudda_data is None:
            if "snudda_data" in self.config_data:
                self.snudda_data = os.path.abspath(self.config_data["snudda_data"])
                self.config_data["snudda_data"] = self.snudda_data  # Make sure abspath is stored
            else:
                snudda_data = ""
        else:
            if "snudda_data" in self.config_data:
                print(f"Warning, overriding snudda_data with {os.path.abspath(self.snudda_data)}")

            self.config_data["snudda_data"] = os.path.abspath(self.snudda_data)

        if not self.snudda_data:
            print(f"Warning, snudda_data is not set!")

    def parse_config(self):

        self.config_data = self.parse_subtree(self.config_data, parent_file=self.config_file)

        self.setup_random_seeds()

    def replace_network_path(self, network_path):
        print(f"Setting {network_path = }")
        self.config_data["network_path"] = network_path

    def skip_item(self, item):

        return isinstance(item, str) and len(item) > 0 and item[0] == '!'

    def get_putative_path(self, putative_file, parent_file=None):

        # This allows us to exclude parsing of certain json files
        if os.path.basename(putative_file) in self.exclude_parse_values:
            return putative_file

        parent_dir = os.path.dirname(parent_file)
        base_dir = os.path.dirname(self.config_file)

        # Version of file name that we check to see if we can find the file
        putative_files = [putative_file,
                          os.path.join(parent_dir, putative_file),
                          os.path.join(base_dir, putative_file),
                          snudda_parse_path(putative_file, snudda_data=self.snudda_data)]
        putative_path = None

        for pf in putative_files:
            if os.path.isfile(pf):
                putative_path = pf
                break

        if putative_path is None:
            raise ValueError(f"File not found {putative_file}")

        return putative_path

    def substitute_json(self, putative_file, key=None, parent_file=None):

        if not (isinstance(putative_file, str) and putative_file.endswith(".json")):
            return putative_file

        # First we look for files in the same directory as the parent directory
        # after that we do SNUDDA_DATA substitution if no match

        putative_path = self.get_putative_path(putative_file=putative_file,
                                               parent_file=parent_file)

        with open(putative_path) as f:
            print(f"Loading {putative_path}")
            sub_tree = json.load(f)

        sub_tree = self.parse_subtree(sub_tree, parent_file=putative_path)

        if key in sub_tree:
            return sub_tree[key]

        return sub_tree

    def parse_subtree(self, config_dict, parent_file=None):

        updated_config = deepcopy(config_dict)

        for key, value in config_dict.items():

            if key in self.exclude_parse_keys:
                continue

            # Keys beginning with ! are excluded
            elif key[0] == "!":
                del updated_config[key]

            elif isinstance(value, dict):
                updated_config[key] = self.parse_subtree(value, parent_file=parent_file)

            elif isinstance(value, list):
                updated_config[key] = [self.substitute_json(x, parent_file=parent_file) for x in value if not self.skip_item(x)]

                # If the first element is a dict, assume all are dict and merge them
                if isinstance(updated_config[key][0], dict):
                    updated_config[key] = dict(ChainMap(*reversed(updated_config[key])))

            else:
                if not self.skip_item(item=value):
                    updated_config[key] = self.substitute_json(value, key, parent_file=parent_file)
                else:
                    del updated_config[key]

        return updated_config

    def setup_random_seeds(self):

        master_seed = None

        if "random_seed" in self.config_data and "master_seed" in self.config_data["random_seed"]:
            master_seed = self.config_data["random_seed"]["master_seed"]

        from snudda.init import SnuddaInit

        rand_seed_dict, init_rng = SnuddaInit.setup_random_seeds(random_seed=master_seed)

        self.config_data["random_seed"] = rand_seed_dict
        self.rng = init_rng

    def write_config(self, output_path=None):

        if output_path is None:
            output_path = os.path.join(self.config_data["network_path"], "network-config.json")

        print(f"Writing to file {output_path}")
        network_dir = os.path.dirname(output_path)
        if network_dir and not os.path.isdir(network_dir):
            os.makedirs(network_dir)

        with open(output_path, "wt") as f:
            json.dump(self.config_data, f, indent=4, cls=NumpyEncoder)


if __name__ == "__main__":

    config_path = "/home/hjorth/HBP/Snudda/new_config/network.json"

    conf = ConfigParser(config_file=config_path, snudda_data="/home/hjorth/HBP/BasalGangliaData/data")
    conf.parse_config()
    conf.write_config()

    # import pdb
    # pdb.set_trace()